#!/usr/bin/perl 
# p8asm version 0.94 -- date: April 21 2010
my $version = '0.94';
my $DEBUG = 0;
#
# (Cross)Assembler for the P800.
# Generated object programs are relocatable by default and must be linked/loaded 
# (possibly with other modules) with p8ldr, which produces an absolute binary 
# (8*8 papertape format) or relocatable binary load module. That binary can then 
# be loaded into the P800 simulater for execution with the simulator 'load' 
# command, or in case of the absolute binary, by booting the simulated P800 
# with that binary.
#
# In order to assemble programs to test an FPGA implementation for a P800,
# version 0.93 adds the possibility to directly generate output in absolute hex 
# format of an absolute source program.
# Output format: <addr> ,<content> <nl>, where <addr> and <content> are 4 digit
# hex numbers. The file extension used for this format is .hex
#
# The assembler makes 2 or 3 passes over the source program:
# 1) if the source program has the .s.. extension (extension which starts with 
#    an s), the the 1st pass performs TAB expansion; the .s.. file is expanded 
#    into a text file (with .txt extension) with the statement fields in fixed 
#    colomns. In case the source file provided to the assembler has the .txt 
#    extension, the source is assumed to be already in the fixed colomn format 
#    and the 1st pass, "the preprocessing pass" is skipped.
#    TAB is either \ or <space>.
# 2) This pass, "pass 1", scans the source file with the .txt extension and 
#    builds the symbol table for the module to define all symbols
# 3) The last pass, "pass 2", scans the .txt source file again, checks the syntax 
#    of each statement and generates the code. Syntax checking and code generation
#    is table driven; both the syntax is scanned and the code is generated
#    with the help of a template available in the table for each operation
#    Additional operations (e.g. the extra operations of the later P800 models)
#    can easily added to the assembler by adding templates for those operations
#    in the table.
#
# Author: Theo Engel
# Info:   Info@theoengel.nl
#
# Usage:  p8asm [-nl|-h] program.sxx|program.txt
#         (the -nl parameter suppresses assembly listing output on the console)
#         (with the -h parameter, .hex output file is generated in case the source
#          program is absolute) 
#
# The assembler is source compatible with the origional P800 assembler, but
# with a number of restrictions. These restrictions are on the todo list. 
# Future improvements of the assembler may remove some of the restrictions. 
# The assembler however, in the stage as is, is suitable for the objective 
# to write programs to test the simulator.
#
# restrictions/to do:
# 1) no conditional assembly
# 2) no FORM, XFORM, STAB, LIST and NLIST speudo operations.
# 3) only assembles a single module (argument 1), so does not assemble
#    a source file with multiple segments separated by :EOS, or multiple source
#    files referenced by multiple arguments.
# 4) DATA speudo operation can only generate a single data item, either a 16 bit
#    value, or (since version 0.4) a n-byte character string 
#    (in case more sequential 16 bit data words are required, then multiple 
#     DATA statements must be written)
# 5) character strings can only be 1 or 2 characters long, except in a DATA
#    statement, where they can have 16-bytes (since version 0.4, which is like
#    in the origional assembler)
# 6) ENTRY and EXTRN speudo operations can have only a single name
#    in the operand field. (in case multiple entry points or externals
#    are required in a module, the multiple ENTRTY/EXTRN statements must be
#    writen)
# 7) Register names are not in the symbol table; so a name can not be
#    equated with one of the predefined register names P,A0..A15.<<<<<<<<[solved in 0.4]
#
# - version 0.5   solves an error in version 0.4 releted with long string
#                 expressions in the DATA speudo operation
#
# - version 0.6   adapted to process P800-C object files (.s2/.s3 files).
#                 (.s2 is the non-optimized opject file and .s3 is optimized)
#                 - support extension .s2/.s3
#                 - allow a leading underscore in a label
#                 - adapted tab setting for .s2/.s3 files to allow for larger labels
#                 - max symbol length is 10 chars (incl underscore), ONLY for .s2/.s3 files
#                 - option to suppress listing output on the console.
#
# - version 0.7  adapted to make the assembler more source compatible with
#                the original P800 assemblers:
#                 - support the character : in symbols/labels (not as 1st character)
#                 - . (dot) is allowed in symbols/labels and instruction codes,
#                   but is not significant.
#                   So the instructions LDKL and LDK.L as well as the
#                   labels L.100 and L100 for example are equivalent.
#                 - condition codes in brance instructions can not only 
#                   be specified as a digit [0-7], but also, as in the later
#                   original assemblers, as a mnemonic:
#                   -------------------------------------------------------------------
#                   CondReg                     Condition
#                   -------------------------------------------------------------------
#                             General   Arith.         Compare            I/O
#                   -------------------------------------------------------------------
#                      =0       (0)   (Z)  Zero     (E)  Equal       (A)  Accepted
#                      =1       (1)   (P)  Pos      (G)  Greater     (R)  Refused
#                      =2       (2)   (N)  Neg      (L)  Less        ---
#                      =3       (3)   (O)  Overfl   ---              (U)  Unknown
#                      #0       (4)   (NZ) Not Zero (NE) Not Equal   (NA) Not Accepted
#                      #1       (5)   (NP) Not Pos  (NG) Not Greater (NR) Not Refused
#                      #2       (6)   (NN) Not Neg  (NL) Not Less    ---
#                      =7       (7)         Unconditional(No condition specified)
#                   -------------------------------------------------------------------
#                 - added a few lacking P856 instructions
#
# - version 0.8 solves an error in the assembly of AB instructions (now uses D iso k)
#
# - version 0.9 adds to expression the hex string in the form X'hhhh' (in addition to /hhhh)
#
# - version 0.91 corrects CCR instruction assembly
#
# - version 0.92 corrects CFI instruction assembly
#
# - verison 0.93 adds .hex output for an absolute source program.
#
# - version 0.94 corrects an error in the bit pattern generated for EXK
#
#
# EQU expressions (in the operand field of the EQU speudo operation) can have
# one of 3 forms (which are not allowed to be mixed in a single expression):
# 1) arithmetic expression of the form: SYM EQU @(+|-)term@(+|-term)...
#    where term is: decimal or hex number or defined abs,rel symbol incl. *
#    So pls. note that the value of all symbols in the expression of an EQU 
#    must be defined in the statements BEFORE the EQU.
# 2) string expression of the form:     SYM EQU '.'|'..'
#    where string is 1 or 2 characters long.
#    If case of 1 character, the string is left padded with a zero byte.
# 3) register expression of the form:   SYM EQU <predefined register name>
# -the value of the expression in the operand field is always a 16 bit value.
# -term in form-1 can not be a quoted string, but can be a symbol defined
#  by form-2
# - term in form-1 cannot be a predefined register, or a register name
#   defined by form-3
#
# DATA expressions (in the operand field of the DATA speudo operation) can
# have one of 2 forms (which are not allowed to be mixed in a single
# expression)
# 1) arithmetic expression of the form: DATA @(+|-)term@(+|-term)...
#    where only 1 external symbol is allowed in an arithmetic expression, and
#    where term is: decimal or hex number or defined abs,rel,extrn symbol incl. *
# 2) string expression of the form:     DATA '<string>'
#    where string is 1,2,..n characters long (n maximum = 48).
#    If n is uneven, the string is right padded with a zero byte.
# - term in form-1 cannot be a quoted string, however in case a symbol is defined
#   by an EQU form-2 statement, then that symbol is allowed in DATA form-1
# - form-1 always results in a 16 bit value
# - register symbols, defined by EQU form-3, are not allowed in a DATA expression.
#
# Long constant expressions (in the operand field of LDKL, ORKL, etc.) can
# have one of 2 forms:
# 1) arithmetic expression of the form: @(+|-)term@(+|-term)...
#    where only 1 external symbol is allowed in an arithmetic expression, and
#    where term is: decimal or hex number or defined abs,rel,extrn symbol incl. *
# 2) string expression of the form:     '.'|'..'
#    where string is 1 or 2 characters long.
#    If case of 1 character, the string is left padded with a zero byte.
# Both forms result in a 16 bit value
#
# Short constant expressions (in the operand field of LDK, ORK, etc.) can
# have one of 2 forms:
# 1) arithmetic expression of the form: @(+|-)term@(+|-term)...
#    where term is: decimal or hex number or defined abs symbol
# 2) string expression of the form:     '.'
#    where string is 1 character long.
# Both forms result in a 8 bit value
#
#
use strict;
use bytes;

# source input fields
#####################
# the source input record is 80 characters maximum (punched card image) and 
# divided in 4 fields: label, operation, operand, comment
#
# default input record layout:
# max symbol length = 6
#***** !**********!!**************!!****!**************************************!
#label operation   operand         comment
#                                       comment
# P800-C input record layout:
# max symbol length = 10
#********* !**********!!**************!!****!**********************************!
#label     operation   operand         comment
#                                       comment
# $tab1=0 $fl1=10 $tab2=11 $fl2=12 $tab3=23 $fl3=21 $tab4=44 $fl4=36
#
# record layout can be changed by adapting the following settings.
# default (p800 assembler)
my $tab1 = 0;               # label field
my $fl1  = 6;               # +1
my $tab2 = 7;               # operation field
my $fl2  = 12;
my $tab3 = 19;              # operand field
# standard operand field
#my $fl3  = 16; 
#my $tab4 = 35;              # comment field
#my $fl4  = 45;
# larger operand field (p800-c code assembler)
my $fl3  = 21; 
my $tab4 = 40;              # comment field
my $fl4  = 40;
my $symlength=6;            # default max symbol length

my $f1;                     # input field1 char 0-5 (l=6)
my $f2;                     # input field2 char 7-18 (l=12)
my $f3;                     # input field3 char 19-34 (l=16)
my $f4;                     # input field4 char 35-79 (l=45)
my $ch;						          # input character
my $line = "";				      # expanded input record
my $cin  = 0;				        # input record char counter
my $cout = 0;				        # expanded record char counter
my $quot = 0;               # in quoted string flag
my $ll;						          # input record length
my $l;                      # temp
my $i;                      # temp
my $conlist=1;              # listing by default to concole as well
my $abshex=0;               # if 1, the abs hex output is generated

my $fi = shift (@ARGV);				
if($fi eq '-nl')
{
	$conlist=0;
	$fi = shift (@ARGV);				# source file
}
if($fi eq '-h')
{
	$abshex=1;
	$fi = shift (@ARGV);				# source file
}	
if ($fi eq "" &&  $fi !~ /\.txt/ && $fi !~ /\.s/)
{
	print "No filename, or wrong extention: $fi\n";
	print "Usage: p8asmx.pl [-nl] <filename>.s..|<filename>.txt\n";
	exit 1;
}

print "P800 cross assembler version $version\n";
my $fn = "";                        # file name
my $ty = "";                        # file type
my $fo = "";						            # object file
my $fl = "";                        # list file

# If the extension is .s.. (or .s2/.s3), pass 0 is executed to expand tabs
# The result is a file with expanded records with a .txt extension
# In case the provided input file has already an .txt extension, pass 0 is
# skipped and pass 1 is entered directly. The .txt file is expected to have
# the fields formatted in the required colums

if( $fi =~ /\.s/)
{
	if( $fi =~ /\.s2/ || $fi =~ /\.s3/)
	{
		# set tab setting for p800-c
		$tab1=0; $fl1=10; $tab2=11; $fl2=12; $tab3=23; $fl3=21; $tab4=44; $fl4=36;
		$symlength = 10;
	}
	($fn,$ty) = split(/\./, $fi);
	$fo = $fn . ".txt";
	if($DEBUG) {print "Expand tabs in $fi to papertape $fo\n";}
	if($DEBUG) {print "Start reading $fi (type is $ty) \n";}
	open (IN,$fi) || die "cannot open inputfile $fi: $!";
	open (OUT,">$fo") || die "cannot open inputfile $fo: $!";
	while(<IN>)
	{
		$line = "";
		$cin  = 0;
		$cout = 0;
		$quot = 0;
		chomp;
		$ll = length($_);
		# expand tabs
		if(($ll > 0) && (substr($_,$cin,1) ne "*"))
		{
			while($cin < $ll)   
			{
				$ch=substr($_,$cin,1);			        # get next char from input record
				$cin=$cin+1;	
				if( $ch eq "\'" && $quot == 0 )
				{
					$quot = 1;							# set the in-quoted-string-flag
				}
				elsif( $ch eq "\'" && $quot == 1 )
				{
					$quot = 0;							# reset the in-quoted-string-flag
				}
				if($quot)
				{
					# in-quoted-string: don't expand tabs
					$line=$line . $ch;
					$cout = $cout + 1;
				}
				else
				{
                    if( $ch =~ /\\/ || $ch =~ /\ /)		    # replace tab by one or more spaces						
					{
						# replace tab by space
						$line=$line . " ";					# expanded input record
						$cout=$cout+1;
						if($cout <= $tab2)
						{
							while($cout<$tab2)
							{
									$line=$line . " ";	# expanded input record
					                $cout=$cout+1;
							}
							next;
						}
						if($cout <= $tab3)
						{
							while($cout<$tab3)
							{
									$line=$line . " ";	# expanded input record
					                $cout=$cout+1;
							}
							next;
						}
						if($cout <= $tab4)
						{
							while($cout<$tab4)
							{
									$line=$line . " ";	# expanded input record
					                $cout=$cout+1;
							}
							next;
						}
					}
					else
					{
						$line=$line . $ch;
						$cout = $cout + 1;
					}
				}
			
			}
		}
		else
		{
			$line = $_;		# comment line
		}
		print OUT $line . "\n";
	}
	close (IN)   || die "can't close $fi: $!";
	close (OUT)  || die "can't close $fo: $!";	
	if($DEBUG) {print "$fi expanded into $fo\n";}
	$fi = $fo;
}

if($ty eq "s2")
{
	$fo = $fn . ".obj";
	$fl = $fn . ".ls2";
}
else
{
	($fn,$ty) = split(/\./, $fi);
	if($abshex)
	{
		$fo = $fn . ".hex";
	}
	else
	{
		$fo = $fn . ".obj";
	}
	$fl = $fn . ".lst";
}

print "Assemble $fi to object papertape $fo and listing $fl\n";
if($DEBUG) {print "Start reading $fi (pass1)\n";}
open (IN,$fi) || die "cannot open inputfile $fi: $!";

my $ln=0;                    # linenumber
my $header="";               # 1st source line is listing header
my $ident=0;                 # 0 if no ident detected, 1 if ident detected
my $nonid=0;                 # set with nonident statements
my $modname="";              # module name, set by ident
my $mode="R";                # default = relocatable mode
my $pass=1;                  # start with pass 1
my $p=0;                     # location counter
my $nerr=0;                  # error counter
my $symbol;                  # scanned symbol (without dots)
my %sym;                     # symbol table (name,value)
my %symtype;                 # symbol type (name,type) R A X Z
#
# symbol types
# a symbol can only defined once and can not be redefined
# symbols are either predefined or are being defined during pass 1
# R  relocatable
# A  absolute
# X  external
# Z  register name
# predefined symbols
$sym{'P'}   = 0x0;	# 0000
$sym{'A0'}  = 0x0;	# 0000
$sym{'A1'}  = 0x2;	# 0010
$sym{'A2'}  = 0x4;	# 0100
$sym{'A3'}  = 0x6;	# 0110
$sym{'A4'}  = 0x8;	# 1000
$sym{'A5'}  = 0xA;	# 1010
$sym{'A6'}  = 0xC;	# 1100
$sym{'A7'}  = 0xE;	# 1110
$sym{'A8'}  = 0x1;	# 0001
$sym{'A9'}  = 0x3;	# 0011
$sym{'A10'} = 0x5;	# 0101
$sym{'A11'} = 0x7;	# 0111
$sym{'A12'} = 0x9;	# 1001
$sym{'A13'} = 0xB;	# 1011
$sym{'A14'} = 0xD;	# 1101
$sym{'A15'} = 0xF;	# 1111

$symtype{'P'}   = "Z";
$symtype{'A0'}  = "Z";
$symtype{'A1'}  = "Z";
$symtype{'A2'}  = "Z";
$symtype{'A3'}  = "Z";
$symtype{'A4'}  = "Z";
$symtype{'A5'}  = "Z";
$symtype{'A6'}  = "Z";
$symtype{'A7'}  = "Z";
$symtype{'A8'}  = "Z";
$symtype{'A9'}  = "Z";
$symtype{'A10'} = "Z";
$symtype{'A11'} = "Z";
$symtype{'A12'} = "Z";
$symtype{'A13'} = "Z";
$symtype{'A14'} = "Z";
$symtype{'A15'} = "Z";

#######################################################################
# expression evaluator
#######################################################################
my $es;					# expression string
my $les;                # expresion string length
my $ei;                 # expresion string index
my $eval;               # expression result
my $etype;              # type result value ("A", "R", "X", "Z" OR "!" if error)
my $exerr;              # 1 if error detected during expression evaluation
my $exsym;              # in case an external is used in the expression
                        # this external must be resolved by the loader
						# keep the name of this external in this variable
						# for the code generator

my %cvdig;              # digit conversion
$cvdig{'0'}=0x0;
$cvdig{'1'}=0x1;
$cvdig{'2'}=0x2;
$cvdig{'3'}=0x3;
$cvdig{'4'}=0x4;
$cvdig{'5'}=0x5;
$cvdig{'6'}=0x6;
$cvdig{'7'}=0x7;
$cvdig{'8'}=0x8;
$cvdig{'9'}=0x9;
$cvdig{'A'}=0xa;
$cvdig{'B'}=0xb;
$cvdig{'C'}=0xc;
$cvdig{'D'}=0xd;
$cvdig{'E'}=0xe;
$cvdig{'F'}=0xf;

#######################################################################
# operand syntax scanner and code generator
#######################################################################
my %I;		# operand syntax template
my %L;		# instruction length (words)
my %G1;		# code generator of word1
my %G2;		# code generator of word2
my %G3;		# code generator of word3

# operand syntax template: 3 fields
#
#	field1	field2	field3
#	 ""		 ""		 ""		empty, or
#	 r1                     register 1, or
#	 r2		 r2             register 2, or
#			 @r2	 @r2    optional register 2, or
#	 r3                     register 3, or
#	 m		 m              mem address, or
#	 k		 k              short constant, or
#	 lk		 lk             long constant, or
#	 ns		 ns             shift count, or
#	 D                      displacement, or
#	 nr                     register count, or
#	 ik                     internal interrupt number, or
#            df             device function bit, or
#            dn   	 dn     device address, or
#    lk1                    1st long constant
#            lk2            2nd long constant

$I{"LD"}      = 'r1,m,@r2'; $L{"LD"}      = 2; $G1{"LD"}      = '10000.r1.10.r2.0'; $G2{"LD"}   = 'm';
$I{"LD*"}     = 'r1,m,@r2'; $L{"LD*"}     = 2; $G1{"LD*"}     = '10000.r1.11.r2.0'; $G2{"LD*"}  = 'm';
$I{"LDR"}     = 'r1,r2';    $L{"LDR"}     = 1; $G1{"LDR"}     = '10000.r1.00.r2.0';
$I{"LDR*"}    = 'r1,r2';    $L{"LDR*"}    = 1; $G1{"LDR*"}    = '10000.r1.01.r2.0';
$I{"LDK"}     = 'r3,k';     $L{"LDK"}     = 1; $G1{"LDK"}     = '00000.r3.k';
$I{"LDKL"}    = 'r1,lk';    $L{"LDKL"}    = 2; $G1{"LDKL"}    = '10000.r1.0100000'; $G2{"LDKL"} = 'lk';
$I{"ST"}      = 'r1,m,@r2'; $L{"ST"}      = 2; $G1{"ST"}      = '10000.r1.10.r2.1'; $G2{"ST"}   = 'm';
$I{"ST*"}     = 'r1,m,@r2'; $L{"ST*"}     = 2; $G1{"ST*"}     = '10000.r1.11.r2.1'; $G2{"ST*"}  = 'm';
$I{"STR"}     = 'r1,r2';    $L{"STR"}     = 1; $G1{"STR"}     = '10000.r1.01.r2.1';
$I{"ML"}      = 'nr,m,@r2'; $L{"ML"}      = 2; $G1{"ML"}      = '10111.nr.10.r2.0'; $G2{"ML"}   = 'm';
$I{"ML*"}     = 'nr,m,@r2'; $L{"ML*"}     = 2; $G1{"ML*"}     = '10111.nr.11.r2.0'; $G2{"ML*"}  = 'm';
$I{"MLK"}     = 'nr';       $L{"MLK"}     = 1; $G1{"MLK"}     = '10111.nr.0100000';
$I{"MLR"}     = 'nr,r2';    $L{"MLR"}     = 1; $G1{"MLR"}     = '10111.nr.01.r2.0';
$I{"MS"}      = 'nr,m,@r2'; $L{"MS"}      = 2; $G1{"MS"}      = '10111.nr.10.r2.1'; $G2{"MS"}   = 'm';
$I{"MS*"}     = 'nr,m,@r2'; $L{"MS*"}     = 2; $G1{"MS*"}     = '10111.nr.11.r2.1'; $G2{"MS*"}  = 'm';
$I{"MSR"}     = 'nr,r2';    $L{"MSR"}     = 1; $G1{"MSR"}     = '10111.nr.01.r2.1';

$I{"ADK"}     = 'r3,k';     $L{"ADK"}     = 1; $G1{"ADK"}     = '00010.r3.k';
$I{"ADKL"}    = 'r1,lk';    $L{"ADKL"}    = 2; $G1{"ADKL"}    = '10010.r1.0100000'; $G2{"ADKL"}  = 'lk';
$I{"ADR"}     = 'r1,r2';    $L{"ADR"}     = 1; $G1{"ADR"}     = '10010.r1.00.r2.0'; 
$I{"ADR*"}    = 'r1,r2';    $L{"ADR*"}    = 1; $G1{"ADR*"}    = '10010.r1.01.r2.0'; 
$I{"ADRS"}    = 'r1,r2';    $L{"ADRS"}    = 1; $G1{"ADRS"}    = '10010.r1.01.r2.1'; 
$I{"AD"}      = 'r1,m,@r2'; $L{"AD"}      = 2; $G1{"AD"}      = '10010.r1.10.r2.0'; $G2{"AD"}    = 'm';
$I{"ADS"}     = 'r1,m,@r2'; $L{"ADS"}     = 2; $G1{"ADS"}     = '10010.r1.10.r2.1'; $G2{"ADS"}   = 'm';
$I{"AD*"}     = 'r1,m,@r2'; $L{"AD*"}     = 2; $G1{"AD*"}     = '10010.r1.11.r2.0'; $G2{"AD*"}   = 'm';
$I{"ADS*"}    = 'r1,m,@r2'; $L{"ADS*"}    = 2; $G1{"ADS*"}    = '10010.r1.11.r2.1'; $G2{"ADS*"}  = 'm';

$I{"SUK"}     = 'r3,k';     $L{"SUK"}     = 1; $G1{"SUK"}     = '00011.r3.k';
$I{"SUKL"}    = 'r1,lk';    $L{"SUKL"}    = 2; $G1{"SUKL"}    = '10011.r1.0100000'; $G2{"SUKL"}  = 'lk';
$I{"SUR"}     = 'r1,r2';    $L{"SUR"}     = 1; $G1{"SUR"}     = '10011.r1.00.r2.0'; 
$I{"SUR*"}    = 'r1,r2';    $L{"SUR*"}    = 1; $G1{"SUR*"}    = '10011.r1.01.r2.0'; 
$I{"SURS"}    = 'r1,r2';    $L{"SURS"}    = 1; $G1{"SURS"}    = '10011.r1.01.r2.1'; 
$I{"SU"}      = 'r1,m,@r2'; $L{"SU"}      = 2; $G1{"SU"}      = '10011.r1.10.r2.0'; $G2{"SU"}    = 'm';
$I{"SUS"}     = 'r1,m,@r2'; $L{"SUS"}     = 2; $G1{"SUS"}     = '10011.r1.10.r2.1'; $G2{"SUS"}   = 'm';
$I{"SU*"}     = 'r1,m,@r2'; $L{"SU*"}     = 2; $G1{"SU*"}     = '10011.r1.11.r2.0'; $G2{"SU*"}   = 'm';
$I{"SUS*"}    = 'r1,m,@r2'; $L{"SUS*"}    = 2; $G1{"SUS*"}    = '10011.r1.11.r2.1'; $G2{"SUS*"}  = 'm';

$I{"ANK"}     = 'r3,k';     $L{"ANK"}     = 1; $G1{"ANK"}     = '00100.r3.k';
$I{"ANKL"}    = 'r1,lk';    $L{"ANKL"}    = 2; $G1{"ANKL"}    = '10100.r1.0100000'; $G2{"ANKL"}  = 'lk';
$I{"ANR"}     = 'r1,r2';    $L{"ANR"}     = 1; $G1{"ANR"}     = '10100.r1.00.r2.0'; 
$I{"ANR*"}    = 'r1,r2';    $L{"ANR*"}    = 1; $G1{"ANR*"}    = '10100.r1.01.r2.0'; 
$I{"ANRS"}    = 'r1,r2';    $L{"ANRS"}    = 1; $G1{"ANRS"}    = '10100.r1.01.r2.1'; 
$I{"AN"}      = 'r1,m,@r2'; $L{"AN"}      = 2; $G1{"AN"}      = '10100.r1.10.r2.0'; $G2{"AN"}    = 'm';
$I{"ANS"}     = 'r1,m,@r2'; $L{"ANS"}     = 2; $G1{"ANS"}     = '10100.r1.10.r2.1'; $G2{"ANS"}   = 'm';
$I{"AN*"}     = 'r1,m,@r2'; $L{"AN*"}     = 2; $G1{"AN*"}     = '10100.r1.11.r2.0'; $G2{"AN*"}   = 'm';
$I{"ANS*"}    = 'r1,m,@r2'; $L{"ANS*"}    = 2; $G1{"ANS*"}    = '10100.r1.11.r2.1'; $G2{"ANS*"}  = 'm';

$I{"TM"}      = 'r1,r2';    $L{"TM"}      = 1; $G1{"TM"}      = '10100.r1.00.r2.1'; 

$I{"ORK"}     = 'r3,k';     $L{"ORK"}     = 1; $G1{"ORK"}     = '00101.r3.k';
$I{"ORKL"}    = 'r1,lk';    $L{"ORKL"}    = 2; $G1{"ORKL"}    = '10101.r1.0100000'; $G2{"ORKL"}  = 'lk';
$I{"ORR"}     = 'r1,r2';    $L{"ORR"}     = 1; $G1{"ORR"}     = '10101.r1.00.r2.0'; 
$I{"ORR*"}    = 'r1,r2';    $L{"ORR*"}    = 1; $G1{"ORR*"}    = '10101.r1.01.r2.0'; 
$I{"ORRS"}    = 'r1,r2';    $L{"ORRS"}    = 1; $G1{"ORRS"}    = '10101.r1.01.r2.1'; 
$I{"OR"}      = 'r1,m,@r2'; $L{"OR"}      = 2; $G1{"OR"}      = '10101.r1.10.r2.0'; $G2{"OR"}    = 'm';
$I{"ORS"}     = 'r1,m,@r2'; $L{"ORS"}     = 2; $G1{"ORS"}     = '10101.r1.10.r2.1'; $G2{"ORS"}   = 'm';
$I{"OR*"}     = 'r1,m,@r2'; $L{"OR*"}     = 2; $G1{"OR*"}     = '10101.r1.11.r2.0'; $G2{"OR*"}   = 'm';
$I{"ORS*"}    = 'r1,m,@r2'; $L{"ORS*"}    = 2; $G1{"ORS*"}    = '10101.r1.11.r2.1'; $G2{"ORS*"}  = 'm';

$I{"XRK"}     = 'r3,k';     $L{"XRK"}     = 1; $G1{"XRK"}     = '00110.r3.k';
$I{"XRKL"}    = 'r1,lk';    $L{"XRKL"}    = 2; $G1{"XRKL"}    = '10110.r1.0100000'; $G2{"XRKL"}  = 'lk';
$I{"XRR"}     = 'r1,r2';    $L{"XRR"}     = 1; $G1{"XRR"}     = '10110.r1.00.r2.0'; 
$I{"XRR*"}    = 'r1,r2';    $L{"XRR*"}    = 1; $G1{"XRR*"}    = '10110.r1.01.r2.0'; 
$I{"XRRS"}    = 'r1,r2';    $L{"XRRS"}    = 1; $G1{"XRRS"}    = '10110.r1.01.r2.1'; 
$I{"XR"}      = 'r1,m,@r2'; $L{"XR"}      = 2; $G1{"XR"}      = '10110.r1.10.r2.0'; $G2{"XR"}    = 'm';
$I{"XRS"}     = 'r1,m,@r2'; $L{"XRS"}     = 2; $G1{"XRS"}     = '10110.r1.10.r2.1'; $G2{"XRS"}   = 'm';
$I{"XR*"}     = 'r1,m,@r2'; $L{"XR*"}     = 2; $G1{"XR*"}     = '10110.r1.11.r2.0'; $G2{"XR*"}   = 'm';
$I{"XRS*"}    = 'r1,m,@r2'; $L{"XRS*"}    = 2; $G1{"XRS*"}    = '10110.r1.11.r2.1'; $G2{"XRS*"}  = 'm';

$I{"TNM"}     = 'r1,r2';    $L{"TNM"}     = 1; $G1{"TNM"}     = '10110.r1.00.r2.1'; 

$I{"ECR"}     = 'r1,r2';    $L{"ECR"}     = 1; $G1{"ECR"}     = '11100.r1.00.r2.0';
$I{"LCK"}     = 'r1,lk';    $L{"LCK"}     = 2; $G1{"LCK"}     = '11100.r1.0100000'; $G2{"LCK"}   = 'lk';
$I{"LCR"}     = 'r1,r2';    $L{"LCR"}     = 1; $G1{"LCR"}     = '11100.r1.01.r2.0'; 
$I{"LC"}      = 'r1,m,@r2'; $L{"LC"}      = 2; $G1{"LC"}      = '11100.r1.10.r2.0'; $G2{"LC"}    = 'm';
$I{"LC*"}     = 'r1,m,@r2'; $L{"LC*"}     = 2; $G1{"LC*"}     = '11100.r1.11.r2.0'; $G2{"LC*"}   = 'm';
$I{"SCR"}     = 'r1,r2';    $L{"SCR"}     = 1; $G1{"SCR"}     = '11100.r1.01.r2.1'; 
$I{"SC"}      = 'r1,m,@r2'; $L{"SC"}      = 2; $G1{"SC"}      = '11100.r1.10.r2.1'; $G2{"SC"}    = 'm';
$I{"SC*"}     = 'r1,m,@r2'; $L{"SC*"}     = 2; $G1{"SC*"}     = '11100.r1.11.r2.1'; $G2{"SC*"}   = 'm';

$I{"CCK"}     = 'r1,lk';    $L{"CCK"}     = 2; $G1{"CCK"}     = '11101.r1.0100001'; $G2{"CCK"}   = 'lk';
$I{"CCR"}     = 'r1,r2';    $L{"CCR"}     = 1; $G1{"CCR"}     = '11101.r1.01.r2.1'; 
$I{"CC"}      = 'r1,m,@r2'; $L{"CC"}      = 2; $G1{"CC"}      = '11101.r1.10.r2.1'; $G2{"CC"}    = 'm';
$I{"CC*"}     = 'r1,m,@r2'; $L{"CC*"}     = 2; $G1{"CC*"}     = '11101.r1.11.r2.1'; $G2{"CC*"}   = 'm';

$I{"CWK"}     = 'r1,lk';    $L{"CWK"}     = 2; $G1{"CWK"}     = '11101.r1.0100000'; $G2{"CWK"}   = 'lk';
$I{"CWR"}     = 'r1,r2';    $L{"CWR"}     = 1; $G1{"CWR"}     = '11101.r1.00.r2.0'; 
$I{"CWR*"}    = 'r1,r2';    $L{"CWR*"}    = 1; $G1{"CWR*"}    = '11101.r1.01.r2.0'; 
$I{"CW"}      = 'r1,m,@r2'; $L{"CW"}      = 2; $G1{"CW"}      = '11101.r1.10.r2.0'; $G2{"CW"}    = 'm';
$I{"CW*"}     = 'r1,m,@r2'; $L{"CW*"}     = 2; $G1{"CW*"}     = '11101.r1.11.r2.0'; $G2{"CW*"}   = 'm';

$I{"C1"}      = 'r1,m,@r2'; $L{"C1"}      = 2; $G1{"C1"}      = '11111.r1.10.r2.0'; $G2{"C1"}    = 'm';
$I{"C1*"}     = 'r1,m,@r2'; $L{"C1*"}     = 2; $G1{"C1*"}     = '11111.r1.11.r2.0'; $G2{"C1*"}   = 'm';
$I{"C1S"}     = 'm,@r2';    $L{"C1S"}     = 2; $G1{"C1S"}     = '11111000010.r2.1'; $G2{"C1S"}   = 'm';
$I{"C1S*"}    = 'm,@r2';    $L{"C1S*"}    = 2; $G1{"C1S*"}    = '11111000011.r2.1'; $G2{"C1S*"}  = 'm';
$I{"C1R"}     = 'r1,r2';    $L{"C1R"}     = 1; $G1{"C1R"}     = '11111.r1.00.r2.0'; 
$I{"C1R*"}    = 'r1,r2';    $L{"C1R*"}    = 1; $G1{"C1R*"}    = '11111.r1.01.r2.0'; 
$I{"C1RS"}    = 'r2';       $L{"C1RS"}    = 1; $G1{"C1RS"}    = '11111000001.r2.1'; 

$I{"NGR"}     = 'r1,r2';    $L{"NGR"}     = 1; $G1{"NGR"}     = '10011.r1.00.r2.1';                         # C2R  not on 855/860
$I{"C2R"}     = 'r2';       $L{"C2R"}     = 1; $G1{"C2R"}     = '10011000001.r2.1';                         # C2RS
$I{"C2"}      = 'm,@r2';    $L{"C2"}      = 2; $G1{"C2"}      = '10011000010.r2.1'; $G2{"C2"}    = 'm';     # C2S
$I{"C2*"}     = 'm,@r2';    $L{"C2*"}     = 2; $G1{"C2*"}     = '10011000011.r2.1'; $G2{"C2*"}   = 'm';     # C2S*

$I{"CMR"}     = 'r2';       $L{"CMR"}     = 1; $G1{"CMR"}     = '10100000001.r2.1';                         # not on 855/860
$I{"CM"}      = 'm,@r2';    $L{"CM"}      = 2; $G1{"CM"}      = '10100000010.r2.1'; $G2{"CM"}    = 'm';     # not on 855/860
$I{"CM*"}     = 'm,@r2';    $L{"CM*"}     = 2; $G1{"CM*"}     = '10100000011.r2.1'; $G2{"CM*"}   = 'm';     # not on 855/860

$I{"MUK"}     = 'lk';       $L{"MUK"}     = 2; $G1{"MUK"}     = '1100000000100000'; $G2{"MUK"}   = 'lk';
$I{"MUR"}     = 'r2';       $L{"MUR"}     = 1; $G1{"MUR"}     = '11000000000.r2.0'; 
$I{"MUR*"}    = 'r2';       $L{"MUR*"}    = 1; $G1{"MUR*"}    = '11000000001.r2.0'; 
$I{"MU"}      = 'm,@r2';    $L{"MU"}      = 2; $G1{"MU"}      = '11000000010.r2.0'; $G2{"MU"}    = 'm';
$I{"MU*"}     = 'm,@r2';    $L{"MU*"}     = 2; $G1{"MU*"}     = '11000000011.r2.0'; $G2{"MO*"}   = 'm';

$I{"DVK"}     = 'lk';       $L{"DVK"}     = 2; $G1{"DVK"}     = '1100100000100000'; $G2{"DVK"}   = 'lk';
$I{"DVR"}     = 'r2';       $L{"DVR"}     = 1; $G1{"DVR"}     = '11001000000.r2.0'; 
$I{"DVR*"}    = 'r2';       $L{"DVR*"}    = 1; $G1{"DVR*"}    = '11001000001.r2.0'; 
$I{"DV"}      = 'm,@r2';    $L{"DV"}      = 2; $G1{"DV"}      = '11001000010.r2.0'; $G2{"DV"}    = 'm';
$I{"DV*"}     = 'm,@r2';    $L{"DV*"}     = 2; $G1{"DV*"}     = '11001000011.r2.0'; $G2{"DV*"}   = 'm';

$I{"DAK"}     = 'lk1,lk2';  $L{"DAK"}     = 3; $G1{"DAK"}     = '1101000000100000'; $G2{"DAK"}   = 'lk1'; $G3{"DAK"}  = 'lk2';
$I{"DAR"}     = 'r2';       $L{"DAR"}     = 1; $G1{"DAR"}     = '11010000000.r2.0'; 
$I{"DAR*"}    = 'r2';       $L{"DAR*"}    = 1; $G1{"DAR*"}    = '11010000001.r2.0'; 
$I{"DA"}      = 'm,@r2';    $L{"DA"}      = 2; $G1{"DA"}      = '11010000010.r2.0'; $G2{"DA"}    = 'm';
$I{"DA*"}     = 'm,@r2';    $L{"DA*"}     = 2; $G1{"DA*"}     = '11010000011.r2.0'; $G2{"DA*"}   = 'm';

$I{"DSK"}     = 'lk1,lk2';  $L{"DSK"}     = 3; $G1{"DSK"}     = '1101100000100000'; $G2{"DSK"}   = 'lk1'; $G3{"DSK"}  = 'lk2';
$I{"DSR"}     = 'r2';       $L{"DSR"}     = 1; $G1{"DSR"}     = '11011000000.r2.0'; 
$I{"DSR*"}    = 'r2';       $L{"DSR*"}    = 1; $G1{"DSR*"}    = '11011000001.r2.0'; 
$I{"DS"}      = 'm,@r2';    $L{"DS"}      = 2; $G1{"DS"}      = '11011000010.r2.0'; $G2{"DS"}    = 'm';
$I{"DS*"}     = 'm,@r2';    $L{"DS*"}     = 2; $G1{"DS*"}     = '11011000011.r2.0'; $G2{"DS*"}   = 'm';

$I{"IMR"}     = 'r2';       $L{"IMR"}     = 1; $G1{"IMR"}     = '10010000001.r2.1';                         # not on 855/860
$I{"IM"}      = 'm,@r2';    $L{"IM"}      = 2; $G1{"IM"}      = '10010000010.r2.1'; $G2{"IM"}    = 'm';     # not on 855/860
$I{"IM*"}     = 'm,@r2';    $L{"IM*"}     = 2; $G1{"IM*"}     = '10010000011.r2.1'; $G2{"IM*"}   = 'm';     # not on 855/860

$I{"RB"}      = 'D';        $L{"RB"}      = 1; $G1{"RB"}      = '01011111.D.0';
$I{"RB(7)"}   = 'D';        $L{"RB(7)"}   = 1; $G1{"RB(7)"}   = '01011111.D.0';
$I{"RB(6)"}   = 'D';        $L{"RB(6)"}   = 1; $G1{"RB(6)"}   = '01011110.D.0';
$I{"RB(5)"}   = 'D';        $L{"RB(5)"}   = 1; $G1{"RB(5)"}   = '01011101.D.0';
$I{"RB(4)"}   = 'D';        $L{"RB(4)"}   = 1; $G1{"RB(4)"}   = '01011100.D.0';
$I{"RB(3)"}   = 'D';        $L{"RB(3)"}   = 1; $G1{"RB(3)"}   = '01011011.D.0';
$I{"RB(2)"}   = 'D';        $L{"RB(2)"}   = 1; $G1{"RB(2)"}   = '01011010.D.0';
$I{"RB(1)"}   = 'D';        $L{"RB(1)"}   = 1; $G1{"RB(1)"}   = '01011001.D.0';
$I{"RB(0)"}   = 'D';        $L{"RB(0)"}   = 1; $G1{"RB(0)"}   = '01011000.D.0';

$I{"RB(NN)"}  = 'D';        $L{"RB(NN)"}  = 1; $G1{"RB(NN)"}  = '01011110.D.0'; #(6)
$I{"RB(NL)"}  = 'D';        $L{"RB(NL)"}  = 1; $G1{"RB(NL)"}  = '01011110.D.0'; #(6)
$I{"RB(NP)"}  = 'D';        $L{"RB(NP)"}  = 1; $G1{"RB(NP)"}  = '01011101.D.0'; #(5)
$I{"RB(NG)"}  = 'D';        $L{"RB(NG)"}  = 1; $G1{"RB(NG)"}  = '01011101.D.0'; #(5)
$I{"RB(NR)"}  = 'D';        $L{"RB(NR)"}  = 1; $G1{"RB(NR)"}  = '01011101.D.0'; #(5)
$I{"RB(NZ)"}  = 'D';        $L{"RB(NZ)"}  = 1; $G1{"RB(NZ)"}  = '01011100.D.0'; #(4)
$I{"RB(NE)"}  = 'D';        $L{"RB(NE)"}  = 1; $G1{"RB(NE)"}  = '01011100.D.0'; #(4)
$I{"RB(NA)"}  = 'D';        $L{"RB(NA)"}  = 1; $G1{"RB(NA)"}  = '01011100.D.0'; #(4)
$I{"RB(O)"}   = 'D';        $L{"RB(O)"}   = 1; $G1{"RB(O)"}   = '01011011.D.0'; #(3)
$I{"RB(U)"}   = 'D';        $L{"RB(U)"}   = 1; $G1{"RB(U)"}   = '01011011.D.0'; #(3)
$I{"RB(N)"}   = 'D';        $L{"RB(N)"}   = 1; $G1{"RB(N)"}   = '01011010.D.0'; #(2)
$I{"RB(L)"}   = 'D';        $L{"RB(L)"}   = 1; $G1{"RB(L)"}   = '01011010.D.0'; #(2)
$I{"RB(P)"}   = 'D';        $L{"RB(P)"}   = 1; $G1{"RB(P)"}   = '01011001.D.0'; #(1)
$I{"RB(G)"}   = 'D';        $L{"RB(G)"}   = 1; $G1{"RB(G)"}   = '01011001.D.0'; #(1)
$I{"RB(R)"}   = 'D';        $L{"RB(R)"}   = 1; $G1{"RB(R)"}   = '01011001.D.0'; #(1)
$I{"RB(Z)"}   = 'D';        $L{"RB(Z)"}   = 1; $G1{"RB(Z)"}   = '01011000.D.0'; #(0)
$I{"RB(E)"}   = 'D';        $L{"RB(E)"}   = 1; $G1{"RB(E)"}   = '01011000.D.0'; #(0)
$I{"RB(A)"}   = 'D';        $L{"RB(A)"}   = 1; $G1{"RB(A)"}   = '01011000.D.0'; #(0)

$I{"RF"}      = 'D';        $L{"RF"}      = 1; $G1{"RF"}      = '01010111.D.0';
$I{"RF(7)"}   = 'D';        $L{"RF(7)"}   = 1; $G1{"RF(7)"}   = '01010111.D.0';
$I{"RF(6)"}   = 'D';        $L{"RF(6)"}   = 1; $G1{"RF(6)"}   = '01010110.D.0';
$I{"RF(5)"}   = 'D';        $L{"RF(5)"}   = 1; $G1{"RF(5)"}   = '01010101.D.0';
$I{"RF(4)"}   = 'D';        $L{"RF(4)"}   = 1; $G1{"RF(4)"}   = '01010100.D.0';
$I{"RF(3)"}   = 'D';        $L{"RF(3)"}   = 1; $G1{"RF(3)"}   = '01010011.D.0';
$I{"RF(2)"}   = 'D';        $L{"RF(2)"}   = 1; $G1{"RF(2)"}   = '01010010.D.0';
$I{"RF(1)"}   = 'D';        $L{"RF(1)"}   = 1; $G1{"RF(1)"}   = '01010001.D.0';
$I{"RF(0)"}   = 'D';        $L{"RF(0)"}   = 1; $G1{"RF(0)"}   = '01010000.D.0';

$I{"RF(NN)"}   = 'D';       $L{"RF(NN)"}   = 1; $G1{"RF(NN)"}   = '01010110.D.0'; #(6)
$I{"RF(NL)"}   = 'D';       $L{"RF(NL)"}   = 1; $G1{"RF(NL)"}   = '01010110.D.0'; #(6)

$I{"RF(NP)"}   = 'D';       $L{"RF(NP)"}   = 1; $G1{"RF(NP)"}   = '01010101.D.0'; #(5)
$I{"RF(NG)"}   = 'D';       $L{"RF(NG)"}   = 1; $G1{"RF(NG)"}   = '01010101.D.0'; #(5)
$I{"RF(NR)"}   = 'D';       $L{"RF(NR)"}   = 1; $G1{"RF(NR)"}   = '01010101.D.0'; #(5)

$I{"RF(NZ)"}   = 'D';       $L{"RF(NZ)"}   = 1; $G1{"RF(NZ)"}   = '01010100.D.0'; #(4)
$I{"RF(NE)"}   = 'D';       $L{"RF(NE)"}   = 1; $G1{"RF(NE)"}   = '01010100.D.0'; #(4)
$I{"RF(NA)"}   = 'D';       $L{"RF(NA)"}   = 1; $G1{"RF(NA)"}   = '01010100.D.0'; #(4)

$I{"RF(O)"}   = 'D';        $L{"RF(O)"}   = 1; $G1{"RF(O)"}   = '01010011.D.0'; #(3)
$I{"RF(U)"}   = 'D';        $L{"RF(U)"}   = 1; $G1{"RF(U)"}   = '01010011.D.0'; #(3)

$I{"RF(N)"}   = 'D';        $L{"RF(N)"}   = 1; $G1{"RF(N)"}   = '01010010.D.0'; #(2)
$I{"RF(L)"}   = 'D';        $L{"RF(L)"}   = 1; $G1{"RF(L)"}   = '01010010.D.0'; #(2)

$I{"RF(P)"}   = 'D';        $L{"RF(P)"}   = 1; $G1{"RF(P)"}   = '01010001.D.0'; #(1)
$I{"RF(G)"}   = 'D';        $L{"RF(G)"}   = 1; $G1{"RF(G)"}   = '01010001.D.0'; #(1)
$I{"RF(R)"}   = 'D';        $L{"RF(R)"}   = 1; $G1{"RF(R)"}   = '01010001.D.0'; #(1)

$I{"RF(Z)"}   = 'D';        $L{"RF(Z)"}   = 1; $G1{"RF(Z)"}   = '01010000.D.0'; #(0)
$I{"RF(E)"}   = 'D';        $L{"RF(E)"}   = 1; $G1{"RF(E)"}   = '01010000.D.0'; #(0)
$I{"RF(A)"}   = 'D';        $L{"RF(A)"}   = 1; $G1{"RF(A)"}   = '01010000.D.0'; #(0)

$I{"AB"}      = 'D';        $L{"AB"}      = 1; $G1{"AB"}      = '00001111.D.0';
$I{"AB(7)"}   = 'D';        $L{"AB(7)"}   = 1; $G1{"AB(7)"}   = '00001111.D.0';
$I{"AB(6)"}   = 'D';        $L{"AB(6)"}   = 1; $G1{"AB(6)"}   = '00001110.D.0';
$I{"AB(5)"}   = 'D';        $L{"AB(5)"}   = 1; $G1{"AB(5)"}   = '00001101.D.0';
$I{"AB(4)"}   = 'D';        $L{"AB(4)"}   = 1; $G1{"AB(4)"}   = '00001100.D.0';
$I{"AB(3)"}   = 'D';        $L{"AB(3)"}   = 1; $G1{"AB(3)"}   = '00001011.D.0';
$I{"AB(2)"}   = 'D';        $L{"AB(2)"}   = 1; $G1{"AB(2)"}   = '00001010.D.0';
$I{"AB(1)"}   = 'D';        $L{"AB(1)"}   = 1; $G1{"AB(1)"}   = '00001001.D.0';
$I{"AB(0)"}   = 'D';        $L{"AB(0)"}   = 1; $G1{"AB(0)"}   = '00001000.D.0';

$I{"AB(NN)"}  = 'D';       $L{"AB(NN)"}   = 1; $G1{"AB(NN)"}  = '00001110.D.0'; #(6)
$I{"AB(NL)"}  = 'D';       $L{"AB(NL)"}   = 1; $G1{"AB(NL)"}  = '00001110.D.0'; #(6)

$I{"AB(NP)"}  = 'D';       $L{"AB(NP)"}   = 1; $G1{"AB(NP)"}  = '00001101.D.0'; #(5)
$I{"AB(NG)"}  = 'D';       $L{"AB(NG)"}   = 1; $G1{"AB(NG)"}  = '00001101.D.0'; #(5)
$I{"AB(NR)"}  = 'D';       $L{"AB(NR)"}   = 1; $G1{"AB(NR)"}  = '00001101.D.0'; #(5)

$I{"AB(NZ)"}  = 'D';       $L{"AB(NZ)"}   = 1; $G1{"AB(NZ)"}  = '00001100.D.0'; #(4)
$I{"AB(NE)"}  = 'D';       $L{"AB(NE)"}   = 1; $G1{"AB(NE)"}  = '00001100.D.0'; #(4)
$I{"AB(NA)"}  = 'D';       $L{"AB(NA)"}   = 1; $G1{"AB(NA)"}  = '00001100.D.0'; #(4)

$I{"AB(O)"}   = 'D';        $L{"AB(O)"}   = 1; $G1{"AB(O)"}   = '00001011.D.0'; #(3)
$I{"AB(U)"}   = 'D';        $L{"AB(U)"}   = 1; $G1{"AB(U)"}   = '00001011.D.0'; #(3)

$I{"AB(N)"}   = 'D';        $L{"AB(N)"}   = 1; $G1{"AB(N)"}   = '00001010.D.0'; #(2)
$I{"AB(L)"}   = 'D';        $L{"AB(L)"}   = 1; $G1{"AB(L)"}   = '00001010.D.0'; #(2)

$I{"AB(P)"}   = 'D';        $L{"AB(P)"}   = 1; $G1{"AB(P)"}   = '00001001.D.0'; #(1)
$I{"AB(G)"}   = 'D';        $L{"AB(G)"}   = 1; $G1{"AB(G)"}   = '00001001.D.0'; #(1)
$I{"AB(R)"}   = 'D';        $L{"AB(R)"}   = 1; $G1{"AB(R)"}   = '00001001.D.0'; #(1)

$I{"AB(Z)"}   = 'D';        $L{"AB(Z)"}   = 1; $G1{"AB(Z)"}   = '00001000.D.0'; #(0)
$I{"AB(E)"}   = 'D';        $L{"AB(E)"}   = 1; $G1{"AB(E)"}   = '00001000.D.0'; #(0)
$I{"AB(A)"}   = 'D';        $L{"AB(A)"}   = 1; $G1{"AB(A)"}   = '00001000.D.0'; #(0)

$I{"ABL"}      = 'lk';      $L{"ABL"}     = 2; $G1{"ABL"}     = '1000111100100000'; $G2{"ABL"}     = 'lk';
$I{"ABL(7)"}   = 'lk';      $L{"ABL(7)"}  = 2; $G1{"ABL(7)"}  = '1000111100100000'; $G2{"ABL(7)"}  = 'lk';
$I{"ABL(6)"}   = 'lk';      $L{"ABL(6)"}  = 2; $G1{"ABL(6)"}  = '1000111000100000'; $G2{"ABL(6)"}  = 'lk';
$I{"ABL(5)"}   = 'lk';      $L{"ABL(5)"}  = 2; $G1{"ABL(5)"}  = '1000110100100000'; $G2{"ABL(5)"}  = 'lk';
$I{"ABL(4)"}   = 'lk';      $L{"ABL(4)"}  = 2; $G1{"ABL(4)"}  = '1000110000100000'; $G2{"ABL(4)"}  = 'lk';
$I{"ABL(3)"}   = 'lk';      $L{"ABL(3)"}  = 2; $G1{"ABL(3)"}  = '1000101100100000'; $G2{"ABL(3)"}  = 'lk';
$I{"ABL(2)"}   = 'lk';      $L{"ABL(2)"}  = 2; $G1{"ABL(2)"}  = '1000101000100000'; $G2{"ABL(2)"}  = 'lk';
$I{"ABL(1)"}   = 'lk';      $L{"ABL(1)"}  = 2; $G1{"ABL(1)"}  = '1000100100100000'; $G2{"ABL(1)"}  = 'lk';
$I{"ABL(0)"}   = 'lk';      $L{"ABL(0)"}  = 2; $G1{"ABL(0)"}  = '1000100000100000'; $G2{"ABL(0)"}  = 'lk';

$I{"ABL(NN)"}   = 'lk';     $L{"ABL(NN)"}  = 2; $G1{"ABL(NN)"}  = '1000111000100000'; $G2{"ABL(NN)"}  = 'lk'; #(6)
$I{"ABL(NL)"}   = 'lk';     $L{"ABL(NL)"}  = 2; $G1{"ABL(NL)"}  = '1000111000100000'; $G2{"ABL(NL)"}  = 'lk'; #(6)

$I{"ABL(NP)"}   = 'lk';     $L{"ABL(NP)"}  = 2; $G1{"ABL(NP)"}  = '1000110100100000'; $G2{"ABL(NP)"}  = 'lk'; #(5)
$I{"ABL(NG)"}   = 'lk';     $L{"ABL(NG)"}  = 2; $G1{"ABL(NG)"}  = '1000110100100000'; $G2{"ABL(NG)"}  = 'lk'; #(5)
$I{"ABL(NR)"}   = 'lk';     $L{"ABL(NR)"}  = 2; $G1{"ABL(NR)"}  = '1000110100100000'; $G2{"ABL(NR)"}  = 'lk'; #(5)

$I{"ABL(NZ)"}   = 'lk';     $L{"ABL(NZ)"}  = 2; $G1{"ABL(NZ)"}  = '1000110000100000'; $G2{"ABL(NZ)"}  = 'lk'; #(4)
$I{"ABL(NE)"}   = 'lk';     $L{"ABL(NE)"}  = 2; $G1{"ABL(NE)"}  = '1000110000100000'; $G2{"ABL(NE)"}  = 'lk'; #(4)
$I{"ABL(NA)"}   = 'lk';     $L{"ABL(NA)"}  = 2; $G1{"ABL(NA)"}  = '1000110000100000'; $G2{"ABL(NA)"}  = 'lk'; #(4)

$I{"ABL(O)"}   = 'lk';      $L{"ABL(O)"}  = 2; $G1{"ABL(O)"}  = '1000101100100000'; $G2{"ABL(O)"}  = 'lk'; #(3)
$I{"ABL(U)"}   = 'lk';      $L{"ABL(U)"}  = 2; $G1{"ABL(U)"}  = '1000101100100000'; $G2{"ABL(U)"}  = 'lk'; #(3)

$I{"ABL(N)"}   = 'lk';      $L{"ABL(N)"}  = 2; $G1{"ABL(N)"}  = '1000101000100000'; $G2{"ABL(N)"}  = 'lk'; #(2)
$I{"ABL(L)"}   = 'lk';      $L{"ABL(L)"}  = 2; $G1{"ABL(L)"}  = '1000101000100000'; $G2{"ABL(L)"}  = 'lk'; #(2)

$I{"ABL(P)"}   = 'lk';      $L{"ABL(P)"}  = 2; $G1{"ABL(P)"}  = '1000100100100000'; $G2{"ABL(P)"}  = 'lk'; #(1)
$I{"ABL(G)"}   = 'lk';      $L{"ABL(G)"}  = 2; $G1{"ABL(G)"}  = '1000100100100000'; $G2{"ABL(G)"}  = 'lk'; #(1)
$I{"ABL(R)"}   = 'lk';      $L{"ABL(R)"}  = 2; $G1{"ABL(R)"}  = '1000100100100000'; $G2{"ABL(R)"}  = 'lk'; #(1)

$I{"ABL(Z)"}   = 'lk';      $L{"ABL(Z)"}  = 2; $G1{"ABL(Z)"}  = '1000100000100000'; $G2{"ABL(Z)"}  = 'lk'; #(0)
$I{"ABL(E)"}   = 'lk';      $L{"ABL(E)"}  = 2; $G1{"ABL(E)"}  = '1000100000100000'; $G2{"ABL(E)"}  = 'lk'; #(0)
$I{"ABL(A)"}   = 'lk';      $L{"ABL(A)"}  = 2; $G1{"ABL(A)"}  = '1000100000100000'; $G2{"ABL(A)"}  = 'lk'; #(0)

$I{"ABI"}     = 'm,@r2';    $L{"ABI"}     = 2; $G1{"ABI"}     = '10001111010.r2.0'; $G2{"ABI"}     = 'm';
$I{"ABI(7)"}  = 'm,@r2';    $L{"ABI(7)"}  = 2; $G1{"ABI(7)"}  = '10001111010.r2.0'; $G2{"ABI(7)"}  = 'm';
$I{"ABI(6)"}  = 'm,@r2';    $L{"ABI(6)"}  = 2; $G1{"ABI(6)"}  = '10001110010.r2.0'; $G2{"ABI(6)"}  = 'm'; 
$I{"ABI(5)"}  = 'm,@r2';    $L{"ABI(5)"}  = 2; $G1{"ABI(5)"}  = '10001101010.r2.0'; $G2{"ABI(5)"}  = 'm'; 
$I{"ABI(4)"}  = 'm,@r2';    $L{"ABI(4)"}  = 2; $G1{"ABI(4)"}  = '10001100010.r2.0'; $G2{"ABI(4)"}  = 'm'; 
$I{"ABI(3)"}  = 'm,@r2';    $L{"ABI(3)"}  = 2; $G1{"ABI(3)"}  = '10001011010.r2.0'; $G2{"ABI(3)"}  = 'm'; 
$I{"ABI(2)"}  = 'm,@r2';    $L{"ABI(2)"}  = 2; $G1{"ABI(2)"}  = '10001010010.r2.0'; $G2{"ABI(2)"}  = 'm'; 
$I{"ABI(1)"}  = 'm,@r2';    $L{"ABI(1)"}  = 2; $G1{"ABI(1)"}  = '10001001010.r2.0'; $G2{"ABI(1)"}  = 'm'; 
$I{"ABI(0)"}  = 'm,@r2';    $L{"ABI(0)"}  = 2; $G1{"ABI(0)"}  = '10001000010.r2.0'; $G2{"ABI(0)"}  = 'm'; 

$I{"ABI(NN)"}  = 'm,@r2';   $L{"ABI(NN)"}  = 2; $G1{"ABI(NN)"}  = '10001110010.r2.0'; $G2{"ABI(NN)"}  = 'm'; #(6)
$I{"ABI(NL)"}  = 'm,@r2';   $L{"ABI(NL)"}  = 2; $G1{"ABI(NL)"}  = '10001110010.r2.0'; $G2{"ABI(NL)"}  = 'm'; #(6)

$I{"ABI(NP)"}  = 'm,@r2';   $L{"ABI(NP)"}  = 2; $G1{"ABI(NP)"}  = '10001101010.r2.0'; $G2{"ABI(NP)"}  = 'm'; #(5)
$I{"ABI(NG)"}  = 'm,@r2';   $L{"ABI(NG)"}  = 2; $G1{"ABI(NG)"}  = '10001101010.r2.0'; $G2{"ABI(NG)"}  = 'm'; #(5)
$I{"ABI(NR)"}  = 'm,@r2';   $L{"ABI(NR)"}  = 2; $G1{"ABI(NR)"}  = '10001101010.r2.0'; $G2{"ABI(NR)"}  = 'm'; #(5)
$I{"ABI(NZ)"}  = 'm,@r2';   $L{"ABI(NZ)"}  = 2; $G1{"ABI(NZ)"}  = '10001100010.r2.0'; $G2{"ABI(NZ)"}  = 'm'; #(4)
$I{"ABI(NE)"}  = 'm,@r2';   $L{"ABI(NE)"}  = 2; $G1{"ABI(NE)"}  = '10001100010.r2.0'; $G2{"ABI(NE)"}  = 'm'; #(4)
$I{"ABI(NA)"}  = 'm,@r2';   $L{"ABI(NA)"}  = 2; $G1{"ABI(NA)"}  = '10001100010.r2.0'; $G2{"ABI(NA)"}  = 'm'; #(4)

$I{"ABI(O)"}  = 'm,@r2';    $L{"ABI(O)"}  = 2; $G1{"ABI(O)"}  = '10001011010.r2.0'; $G2{"ABI(O)"}  = 'm'; #(3)
$I{"ABI(U)"}  = 'm,@r2';    $L{"ABI(U)"}  = 2; $G1{"ABI(U)"}  = '10001011010.r2.0'; $G2{"ABI(U)"}  = 'm'; #(3)

$I{"ABI(N)"}  = 'm,@r2';    $L{"ABI(N)"}  = 2; $G1{"ABI(N)"}  = '10001010010.r2.0'; $G2{"ABI(N)"}  = 'm'; #(2)
$I{"ABI(L)"}  = 'm,@r2';    $L{"ABI(L)"}  = 2; $G1{"ABI(L)"}  = '10001010010.r2.0'; $G2{"ABI(L)"}  = 'm'; #(2)

$I{"ABI(P)"}  = 'm,@r2';    $L{"ABI(P)"}  = 2; $G1{"ABI(P)"}  = '10001001010.r2.0'; $G2{"ABI(P)"}  = 'm'; #(1)
$I{"ABI(G)"}  = 'm,@r2';    $L{"ABI(G)"}  = 2; $G1{"ABI(G)"}  = '10001001010.r2.0'; $G2{"ABI(G)"}  = 'm'; #(1)
$I{"ABI(R)"}  = 'm,@r2';    $L{"ABI(R)"}  = 2; $G1{"ABI(R)"}  = '10001001010.r2.0'; $G2{"ABI(R)"}  = 'm'; #(1)

$I{"ABI(Z)"}  = 'm,@r2';    $L{"ABI(Z)"}  = 2; $G1{"ABI(Z)"}  = '10001000010.r2.0'; $G2{"ABI(Z)"}  = 'm'; #(0)
$I{"ABI(E)"}  = 'm,@r2';    $L{"ABI(E)"}  = 2; $G1{"ABI(E)"}  = '10001000010.r2.0'; $G2{"ABI(E)"}  = 'm'; #(0)
$I{"ABI(A)"}  = 'm,@r2';    $L{"ABI(A)"}  = 2; $G1{"ABI(A)"}  = '10001000010.r2.0'; $G2{"ABI(A)"}  = 'm'; #(0)

$I{"ABI*"}    = 'm,@r2';    $L{"ABI*"}    = 2; $G1{"ABI*"}    = '10001111011.r2.0'; $G2{"ABI*"}     = 'm';
$I{"ABI(7)*"} = 'm,@r2';    $L{"ABI(7)*"} = 2; $G1{"ABI(7)*"} = '10001111011.r2.0'; $G2{"ABI(7)*"}  = 'm';
$I{"ABI(6)*"} = 'm,@r2';    $L{"ABI(6)*"} = 2; $G1{"ABI(6)*"} = '10001110011.r2.0'; $G2{"ABI(6)*"}  = 'm';
$I{"ABI(5)*"} = 'm,@r2';    $L{"ABI(5)*"} = 2; $G1{"ABI(5)*"} = '10001101011.r2.0'; $G2{"ABI(5)*"}  = 'm';
$I{"ABI(4)*"} = 'm,@r2';    $L{"ABI(4)*"} = 2; $G1{"ABI(4)*"} = '10001100011.r2.0'; $G2{"ABI(4)*"}  = 'm';
$I{"ABI(3)*"} = 'm,@r2';    $L{"ABI(3)*"} = 2; $G1{"ABI(3)*"} = '10001011011.r2.0'; $G2{"ABI(3)*"}  = 'm';
$I{"ABI(2)*"} = 'm,@r2';    $L{"ABI(2)*"} = 2; $G1{"ABI(2)*"} = '10001010011.r2.0'; $G2{"ABI(2)*"}  = 'm';
$I{"ABI(1)*"} = 'm,@r2';    $L{"ABI(1)*"} = 2; $G1{"ABI(1)*"} = '10001001011.r2.0'; $G2{"ABI(1)*"}  = 'm';
$I{"ABI(0)*"} = 'm,@r2';    $L{"ABI(0)*"} = 2; $G1{"ABI(0)*"} = '10001000011.r2.0'; $G2{"ABI(0)*"}  = 'm';

$I{"ABI(NN)*"} = 'm,@r2';   $L{"ABI(NN)*"} = 2; $G1{"ABI(NN)*"} = '10001110011.r2.0'; $G2{"ABI(NN)*"}  = 'm'; #(6)
$I{"ABI(NL)*"} = 'm,@r2';   $L{"ABI(NL)*"} = 2; $G1{"ABI(NL)*"} = '10001110011.r2.0'; $G2{"ABI(NL)*"}  = 'm'; #(6)

$I{"ABI(NP)*"} = 'm,@r2';   $L{"ABI(NP)*"} = 2; $G1{"ABI(NP)*"} = '10001101011.r2.0'; $G2{"ABI(NP)*"}  = 'm'; #(5)
$I{"ABI(NG)*"} = 'm,@r2';   $L{"ABI(NG)*"} = 2; $G1{"ABI(NG)*"} = '10001101011.r2.0'; $G2{"ABI(NG)*"}  = 'm'; #(5)
$I{"ABI(NR)*"} = 'm,@r2';   $L{"ABI(NR)*"} = 2; $G1{"ABI(NR)*"} = '10001101011.r2.0'; $G2{"ABI(NR)*"}  = 'm'; #(5)

$I{"ABI(NZ)*"} = 'm,@r2';   $L{"ABI(NZ)*"} = 2; $G1{"ABI(NZ)*"} = '10001100011.r2.0'; $G2{"ABI(NZ)*"}  = 'm'; #(4)
$I{"ABI(NE)*"} = 'm,@r2';   $L{"ABI(NE)*"} = 2; $G1{"ABI(NE)*"} = '10001100011.r2.0'; $G2{"ABI(NE)*"}  = 'm'; #(4)
$I{"ABI(NA)*"} = 'm,@r2';   $L{"ABI(NA)*"} = 2; $G1{"ABI(NA)*"} = '10001100011.r2.0'; $G2{"ABI(NA)*"}  = 'm'; #(4)

$I{"ABI(O)*"} = 'm,@r2';    $L{"ABI(O)*"} = 2; $G1{"ABI(O)*"} = '10001011011.r2.0'; $G2{"ABI(O)*"}  = 'm'; #(3)
$I{"ABI(U)*"} = 'm,@r2';    $L{"ABI(U)*"} = 2; $G1{"ABI(U)*"} = '10001011011.r2.0'; $G2{"ABI(U)*"}  = 'm'; #(3)

$I{"ABI(N)*"} = 'm,@r2';    $L{"ABI(N)*"} = 2; $G1{"ABI(N)*"} = '10001010011.r2.0'; $G2{"ABI(N)*"}  = 'm'; #(2)
$I{"ABI(L)*"} = 'm,@r2';    $L{"ABI(L)*"} = 2; $G1{"ABI(L)*"} = '10001010011.r2.0'; $G2{"ABI(L)*"}  = 'm'; #(2)

$I{"ABI(P)*"} = 'm,@r2';    $L{"ABI(P)*"} = 2; $G1{"ABI(P)*"} = '10001001011.r2.0'; $G2{"ABI(P)*"}  = 'm'; #(1)
$I{"ABI(G)*"} = 'm,@r2';    $L{"ABI(G)*"} = 2; $G1{"ABI(G)*"} = '10001001011.r2.0'; $G2{"ABI(G)*"}  = 'm'; #(1)
$I{"ABI(R)*"} = 'm,@r2';    $L{"ABI(R)*"} = 2; $G1{"ABI(R)*"} = '10001001011.r2.0'; $G2{"ABI(R)*"}  = 'm'; #(1)

$I{"ABI(Z)*"} = 'm,@r2';    $L{"ABI(Z)*"} = 2; $G1{"ABI(Z)*"} = '10001000011.r2.0'; $G2{"ABI(Z)*"}  = 'm'; #(0)
$I{"ABI(E)*"} = 'm,@r2';    $L{"ABI(E)*"} = 2; $G1{"ABI(E)*"} = '10001000011.r2.0'; $G2{"ABI(E)*"}  = 'm'; #(0)
$I{"ABI(A)*"} = 'm,@r2';    $L{"ABI(A)*"} = 2; $G1{"ABI(A)*"} = '10001000011.r2.0'; $G2{"ABI(A)*"}  = 'm'; #(0)

$I{"ABR"}     = 'r2';       $L{"ABR"}     = 1; $G1{"ABR"}     = '10001111000.r2.0';
$I{"ABR(7)"}  = 'r2';       $L{"ABR(7)"}  = 1; $G1{"ABR(7)"}  = '10001111000.r2.0';
$I{"ABR(6)"}  = 'r2';       $L{"ABR(6)"}  = 1; $G1{"ABR(6)"}  = '10001110000.r2.0';
$I{"ABR(5)"}  = 'r2';       $L{"ABR(5)"}  = 1; $G1{"ABR(5)"}  = '10001101000.r2.0';
$I{"ABR(4)"}  = 'r2';       $L{"ABR(4)"}  = 1; $G1{"ABR(4)"}  = '10001100000.r2.0';
$I{"ABR(3)"}  = 'r2';       $L{"ABR(3)"}  = 1; $G1{"ABR(3)"}  = '10001011000.r2.0';
$I{"ABR(2)"}  = 'r2';       $L{"ABR(2)"}  = 1; $G1{"ABR(2)"}  = '10001010000.r2.0';
$I{"ABR(1)"}  = 'r2';       $L{"ABR(1)"}  = 1; $G1{"ABR(1)"}  = '10001001000.r2.0';
$I{"ABR(0)"}  = 'r2';       $L{"ABR(0)"}  = 1; $G1{"ABR(0)"}  = '10001000000.r2.0';

$I{"ABR(NN)"}  = 'r2';      $L{"ABR(NN)"}  = 1; $G1{"ABR(NN)"}  = '10001110000.r2.0'; #(6)
$I{"ABR(NL)"}  = 'r2';      $L{"ABR(NL)"}  = 1; $G1{"ABR(NL)"}  = '10001110000.r2.0'; #(6)

$I{"ABR(NP)"}  = 'r2';      $L{"ABR(NP)"}  = 1; $G1{"ABR(NP)"}  = '10001101000.r2.0'; #(5)
$I{"ABR(NG)"}  = 'r2';      $L{"ABR(NG)"}  = 1; $G1{"ABR(NG)"}  = '10001101000.r2.0'; #(5)
$I{"ABR(NR)"}  = 'r2';      $L{"ABR(NR)"}  = 1; $G1{"ABR(NR)"}  = '10001101000.r2.0'; #(5)

$I{"ABR(NZ)"}  = 'r2';      $L{"ABR(NZ)"}  = 1; $G1{"ABR(NZ)"}  = '10001100000.r2.0'; #(4)
$I{"ABR(NE)"}  = 'r2';      $L{"ABR(NE)"}  = 1; $G1{"ABR(NE)"}  = '10001100000.r2.0'; #(4)
$I{"ABR(NA)"}  = 'r2';      $L{"ABR(NA)"}  = 1; $G1{"ABR(NA)"}  = '10001100000.r2.0'; #(4)

$I{"ABR(O)"}  = 'r2';       $L{"ABR(O)"}  = 1; $G1{"ABR(O)"}  = '10001011000.r2.0'; #(3)
$I{"ABR(U)"}  = 'r2';       $L{"ABR(U)"}  = 1; $G1{"ABR(U)"}  = '10001011000.r2.0'; #(3)

$I{"ABR(N)"}  = 'r2';       $L{"ABR(N)"}  = 1; $G1{"ABR(N)"}  = '10001010000.r2.0'; #(2)
$I{"ABR(L)"}  = 'r2';       $L{"ABR(L)"}  = 1; $G1{"ABR(L)"}  = '10001010000.r2.0'; #(2)

$I{"ABR(P)"}  = 'r2';       $L{"ABR(P)"}  = 1; $G1{"ABR(P)"}  = '10001001000.r2.0'; #(1)
$I{"ABR(G)"}  = 'r2';       $L{"ABR(G)"}  = 1; $G1{"ABR(G)"}  = '10001001000.r2.0'; #(1)
$I{"ABR(R)"}  = 'r2';       $L{"ABR(R)"}  = 1; $G1{"ABR(R)"}  = '10001001000.r2.0'; #(1)

$I{"ABR(Z)"}  = 'r2';       $L{"ABR(Z)"}  = 1; $G1{"ABR(Z)"}  = '10001000000.r2.0'; #(0)
$I{"ABR(E)"}  = 'r2';       $L{"ABR(E)"}  = 1; $G1{"ABR(E)"}  = '10001000000.r2.0'; #(0)
$I{"ABR(A)"}  = 'r2';       $L{"ABR(A)"}  = 1; $G1{"ABR(A)"}  = '10001000000.r2.0'; #(0)

$I{"ABR*"}    = 'r2';       $L{"ABR*"}    = 1; $G1{"ABR*"}    = '10001111001.r2.0';
$I{"ABR(7)*"} = 'r2';       $L{"ABR(7)*"} = 1; $G1{"ABR(7)*"} = '10001111001.r2.0';
$I{"ABR(6)*"} = 'r2';       $L{"ABR(6)*"} = 1; $G1{"ABR(6)*"} = '10001110001.r2.0';
$I{"ABR(5)*"} = 'r2';       $L{"ABR(5)*"} = 1; $G1{"ABR(5)*"} = '10001101001.r2.0';
$I{"ABR(4)*"} = 'r2';       $L{"ABR(4)*"} = 1; $G1{"ABR(4)*"} = '10001100001.r2.0';
$I{"ABR(3)*"} = 'r2';       $L{"ABR(3)*"} = 1; $G1{"ABR(3)*"} = '10001011001.r2.0';
$I{"ABR(2)*"} = 'r2';       $L{"ABR(2)*"} = 1; $G1{"ABR(2)*"} = '10001010001.r2.0';
$I{"ABR(1)*"} = 'r2';       $L{"ABR(1)*"} = 1; $G1{"ABR(1)*"} = '10001001001.r2.0';
$I{"ABR(0)*"} = 'r2';       $L{"ABR(0)*"} = 1; $G1{"ABR(0)*"} = '10001000001.r2.0';

$I{"ABR(NN)*"} = 'r2';      $L{"ABR(NN)*"} = 1; $G1{"ABR(NN)*"} = '10001110001.r2.0'; #(6)
$I{"ABR(NL)*"} = 'r2';      $L{"ABR(NL)*"} = 1; $G1{"ABR(NL)*"} = '10001110001.r2.0'; #(6)

$I{"ABR(NP)*"} = 'r2';      $L{"ABR(NP)*"} = 1; $G1{"ABR(NP)*"} = '10001101001.r2.0'; #(5)
$I{"ABR(NG)*"} = 'r2';      $L{"ABR(NG)*"} = 1; $G1{"ABR(NG)*"} = '10001101001.r2.0'; #(5)
$I{"ABR(NR)*"} = 'r2';      $L{"ABR(NR)*"} = 1; $G1{"ABR(NR)*"} = '10001101001.r2.0'; #(5)

$I{"ABR(NZ)*"} = 'r2';      $L{"ABR(NZ)*"} = 1; $G1{"ABR(NZ)*"} = '10001100001.r2.0'; #(4)
$I{"ABR(NE)*"} = 'r2';      $L{"ABR(NE)*"} = 1; $G1{"ABR(NE)*"} = '10001100001.r2.0'; #(4)
$I{"ABR(NA)*"} = 'r2';      $L{"ABR(NA)*"} = 1; $G1{"ABR(NA)*"} = '10001100001.r2.0'; #(4)

$I{"ABR(O)*"} = 'r2';       $L{"ABR(O)*"} = 1; $G1{"ABR(O)*"} = '10001011001.r2.0'; #(3)
$I{"ABR(U)*"} = 'r2';       $L{"ABR(U)*"} = 1; $G1{"ABR(U)*"} = '10001011001.r2.0'; #(3)

$I{"ABR(N)*"} = 'r2';       $L{"ABR(N)*"} = 1; $G1{"ABR(N)*"} = '10001010001.r2.0'; #(2)
$I{"ABR(L)*"} = 'r2';       $L{"ABR(L)*"} = 1; $G1{"ABR(L)*"} = '10001010001.r2.0'; #(2)

$I{"ABR(P)*"} = 'r2';       $L{"ABR(P)*"} = 1; $G1{"ABR(P)*"} = '10001001001.r2.0'; #(1)
$I{"ABR(G)*"} = 'r2';       $L{"ABR(G)*"} = 1; $G1{"ABR(G)*"} = '10001001001.r2.0'; #(1)
$I{"ABR(R)*"} = 'r2';       $L{"ABR(R)*"} = 1; $G1{"ABR(R)*"} = '10001001001.r2.0'; #(1)

$I{"ABR(Z)*"} = 'r2';       $L{"ABR(Z)*"} = 1; $G1{"ABR(Z)*"} = '10001000001.r2.0'; #(0)
$I{"ABR(E)*"} = 'r2';       $L{"ABR(E)*"} = 1; $G1{"ABR(E)*"} = '10001000001.r2.0'; #(0)
$I{"ABR(A)*"} = 'r2';       $L{"ABR(A)*"} = 1; $G1{"ABR(A)*"} = '10001000001.r2.0'; #(0)

$I{"CF"}      = 'r1,lk';    $L{"CF"}      = 2; $G1{"CF"}      = '11110.r1.0100001'; $G2{"CF"}    = 'lk';
$I{"CFI"}     = 'r1,m,@r2'; $L{"CFI"}     = 2; $G1{"CFI"}     = '11110.r1.10.r2.1'; $G2{"CFI"}   = 'm';
$I{"CFI*"}    = 'r1,m,@r2'; $L{"CFI*"}    = 2; $G1{"CFI*"}    = '11110.r1.11.r2.1'; $G2{"CFI*"}  = 'm';
$I{"CFR"}     = 'r1,r2';    $L{"CFR"}     = 1; $G1{"CFR"}     = '11110.r1.00.r2.1'; 
$I{"CFR*"}    = 'r1,r2';    $L{"CFR*"}    = 1; $G1{"CFR*"}    = '11110.r1.01.r2.1';
$I{"RTN"}     = 'r2';       $L{"RTN"}     = 1; $G1{"RTN"}     = '11110000001.r2.0'; 
$I{"EXR"}     = 'r2';       $L{"EXR"}     = 1; $G1{"EXR"}     = '11110000000.r2.1';                     # not on 855/860
$I{"EXR*"}    = 'r2';       $L{"EXR*"}    = 1; $G1{"EXR*"}    = '11110000001.r2.1';                     # not on 855/860
$I{"EXK"}     = 'lk';       $L{"EXK"}     = 2; $G1{"EXK"}     = '1111000000100001'; $G2{"EXK"}  = 'lk'; # not on 855/860
$I{"EX"}      = 'm,@r2';    $L{"EX"}      = 2; $G1{"EX"}      = '11110000010.r2.1'; $G2{"EX"}   = 'm';  # not on 855/860
$I{"EX*"}     = 'm,@r2';    $L{"EX*"}     = 2; $G1{"EX*"}     = '11110000011.r2.1'; $G2{"EX*"}  = 'm';  # not on 855/860

$I{"SLA"}     = 'r3,ns';    $L{"SLA"}     = 1; $G1{"SLA"}     = '00111.r3.000.ns';
$I{"SRA"}     = 'r3,ns';    $L{"SRA"}     = 1; $G1{"SRA"}     = '00111.r3.001.ns';
$I{"SLL"}     = 'r3,ns';    $L{"SLL"}     = 1; $G1{"SLL"}     = '00111.r3.010.ns';
$I{"SRL"}     = 'r3,ns';    $L{"SRL"}     = 1; $G1{"SRL"}     = '00111.r3.011.ns';
$I{"SLC"}     = 'r3,ns';    $L{"SLC"}     = 1; $G1{"SLC"}     = '00111.r3.110.ns';
$I{"SRC"}     = 'r3,ns';    $L{"SRC"}     = 1; $G1{"SRC"}     = '00111.r3.111.ns';
$I{"SLN"}     = 'r3,r2';    $L{"SLN"}     = 1; $G1{"SLN"}     = '00111.r3.100.r2.0';
$I{"SRN"}     = 'r3,r2';    $L{"SRN"}     = 1; $G1{"SRN"}     = '00111.r3.101.r2.0';
$I{"DLA"}     = 'ns';       $L{"DLA"}     = 1; $G1{"DLA"}     = '00111000000.ns';
$I{"DRA"}     = 'ns';       $L{"DRA"}     = 1; $G1{"DRA"}     = '00111000001.ns';
$I{"DLL"}     = 'ns';       $L{"DLL"}     = 1; $G1{"DLL"}     = '00111000010.ns';
$I{"DRL"}     = 'ns';       $L{"DRL"}     = 1; $G1{"DRL"}     = '00111000011.ns';
$I{"DLC"}     = 'ns';       $L{"DLC"}     = 1; $G1{"DLC"}     = '00111000110.ns';
$I{"DRC"}     = 'ns';       $L{"DRC"}     = 1; $G1{"DRC"}     = '00111000111.ns';
$I{"DLN"}     = 'ns';       $L{"DLN"}     = 1; $G1{"DLN"}     = '00111000100.r2.0';
$I{"DRN"}     = 'ns';       $L{"DRN"}     = 1; $G1{"DRN"}     = '00111000101.r2.0';

$I{"HLT"}     = '';         $L{"HLT"}     = 1; $G1{"HLT"}     = '0010000001111111';
$I{"INH"}     = '';         $L{"INH"}     = 1; $G1{"INH"}     = '0010000010111111';
$I{"ENB"}     = '';         $L{"ENB"}     = 1; $G1{"ENB"}     = '0010100001000000';
$I{"LKM"}     = '';         $L{"LKM"}     = 1; $G1{"LKM"}     = '0010100000000100';
$I{"RIT"}     = 'ik';       $L{"RIT"}     = 1; $G1{"RIT"}     = '0010000011.ik.1';          
$I{"SMD"}     = '';         $L{"SMD"}     = 1; $G1{"SMD"}     = '0010100000000001';
$I{"RCA"}     = 'r3';       $L{"RCA"}     = 1; $G1{"RCA"}     = '01001.r3.01000000';        # not on 852/856/etc (read channel address-midb)
$I{"RIL"}     = 'r3';       $L{"RIL"}     = 1; $G1{"RIL"}     = '01001.r3.00000000';        # not on 852/856/etc (read int lines)
$I{"WIM"}     = 'r3';       $L{"WIM"}     = 1; $G1{"WIM"}     = '01000.r3.00000000';        # not on 852/856/etc (write int mask)
$I{"WMP"}     = 'r3';       $L{"WMP"}     = 1; $G1{"WMP"}     = '01000.r3.01000000';        # not on 852/856/etc (write mask protection)
$I{"WM2"}     = 'r3';       $L{"WM2"}     = 1; $G1{"WM2"}     = '01000.r3.11000000';        # not on 852/856/etc (write mask protection 2)

$I{"CIO"}     = 'r3,df,dn'; $L{"CIO"}     = 1; $G1{"CIO"}     = '01000.r3.1.df.dn';
$I{"OTR"}     = 'r3,df,dn'; $L{"OTR"}     = 1; $G1{"OTR"}     = '01000.r3.0.df.dn';
$I{"INR"}     = 'r3,df,dn'; $L{"INR"}     = 1; $G1{"INR"}     = '01001.r3.0.df.dn';
$I{"SST"}     = 'r3,dn';    $L{"SST"}     = 1; $G1{"SST"}     = '01001.r3.11.dn';
$I{"TST"}     = 'r3,dn';    $L{"TST"}     = 1; $G1{"TST"}     = '01001.r3.10.dn';

$I{"WER"}     = 'r3,k';     $L{"WER"}     = 1; $G1{"WER"}     = '01110.r3.k';               # not on 855/860
$I{"RER"}     = 'r3,k';     $L{"RER"}     = 1; $G1{"RER"}     = '01111.r3.k';               # not on 855/860

my $r1;			# r1 field (4 bit)
my $r2;         # r2 field (4 bit)
my $r3;         # r3 field (3 bit)
my $m;          # address field (16 bit)
my $oty;        # operand type (A, R, X)
my $ns;         # shift count field (5 bit)
my $nr;         # register count field (4 bit)
my $k;          # short constant (8 bit)
my $lk;         # long constant (16 bit)
my $lk1;        # long constant 1 (16 bit)
my $lk2;        # long constant 2 (16 bit)
my $oty1;       # 1st operand (lk1) type (A, R, X)
my $oty2;       # 2nd operand (lk2) type (A, R, X)
my $exsym1;     # 1st external symbol (lk1) (if any)
my $exsym2;     # 2nd external symbol (lk2) (if any)
my $ik;         # internal interrupt number (5 bit)
my $D;          # displacement field (7 bit)
my $df;         # device function bit (eg start/stop bit CIO) (1 bit)
my $dn;         # device number (6 bit)

######################## bin code generation ##########################
my $som     = 0201;	#start of codeblock
my $eom     = 0203;	#end   of codeblock
my $space   = 0x20; #space
my $rub     = 0xFF; #rubout
my $prl;			#program length in words
#######################################################################
# ASCII CONVERSION TABLE
my %ch;
$ch{' '}=0x20;
$ch{'!'}=0x21;
$ch{"\""}=0x22;
$ch{"\#"}=0x23;
$ch{'$'}=0x24;
$ch{'%'}=0x25;
$ch{'&'}=0x26;
$ch{"\'"}=0x27;
$ch{'('}=0x28;
$ch{')'}=0x29;
$ch{'*'}=0x2A;
$ch{'+'}=0x2B;
$ch{','}=0x2C;
$ch{'-'}=0x2D;
$ch{'.'}=0x2E;
$ch{"\/"}=0x2F;
$ch{'0'}=0x30;
$ch{'1'}=0x31;
$ch{'2'}=0x32;
$ch{'3'}=0x33;
$ch{'4'}=0x34;
$ch{'5'}=0x35;
$ch{'6'}=0x36;
$ch{'7'}=0x37;
$ch{'8'}=0x38;
$ch{'9'}=0x39;
$ch{"\:"}=0x3A;
$ch{"\;"}=0x3B;
$ch{'<'}=0x3C;
$ch{'='}=0x3D;
$ch{'>'}=0x3E;
$ch{'?'}=0x3F;
$ch{'@'}=0x40;
$ch{'A'}=0x41;
$ch{'B'}=0x42;
$ch{'C'}=0x43;
$ch{'D'}=0x44;
$ch{'E'}=0x45;
$ch{'F'}=0x46;
$ch{'G'}=0x47;
$ch{'H'}=0x48;
$ch{'I'}=0x49;
$ch{'J'}=0x4A;
$ch{'K'}=0x4B;
$ch{'L'}=0x4C;
$ch{'M'}=0x4D;
$ch{'N'}=0x4E;
$ch{'O'}=0x4F;
$ch{'P'}=0x50;
$ch{'Q'}=0x51;
$ch{'R'}=0x52;
$ch{'S'}=0x53;
$ch{'T'}=0x54;
$ch{'U'}=0x55;
$ch{'V'}=0x56;
$ch{'W'}=0x57;
$ch{'X'}=0x58;
$ch{'Y'}=0x59;
$ch{'Z'}=0x5A;
$ch{'['}=0x5B;
$ch{"\\"}=0x5C;
$ch{']'}=0x5D;
$ch{'^'}=0x5E;
$ch{'_'}=0x5F;

########################        PASS 1       ##########################
# input is a text file (extension is .txt) with all tabs are expanded
while(<IN>)
{
	chomp;
	$line = $_;
	if(ord(substr($_,0,1)) == 0)
	{
		# strip leader
		$i=0;
		$l=length($_);
		while((ord(substr($_,$i,1)) == 0) && $i < $l)
		{
			$i=$i+1;
		}
		$line = substr($_,$i,$l-$i);
	}
	$ln=$ln+1;			                   # incr line number
    if($ln == 1)
	{
		$header = $line;
	}
    next if (substr($_,0,1) eq '*');	   # ignore comment lines
    $f1=substr($line,$tab1,$fl1);          # label
	$f1=rtrim($f1);
    $f2=substr($line,$tab2,$fl2);          # operation
	$f2=rtrim($f2);
	$f2=~ tr/\.//d;                        # delete dots from opcode
    $f3=substr($line,$tab3,$fl3);          # operand
	$f3=rtrim($f3);
	
	# print "fields.$f1.$f2.$f3.\n";
	
	# A label is not allowed for the following directives
	if($f2 eq 'IDENT')
	{
		error("Label not allowed in IDENT") if(label());
		if($ident == 0 && $nonid == 0)
		{
			if(symbol($f3))
			{
				$modname=$f3;
				$ident = 1;		# ident detected as 1st statement
			}
			else
			{
				error("Invalid module name");
			}
			next;
		}
		elsif($ident == 1)
		{
			error("Multiple IDENT statements are not allowed");
			next;
		}
		else
		{
			error("IDENT not detected as 1st statement");
			next;
		}
	}	
	if($f2 eq 'ENTRY')
	{
		$nonid=1;
		error("Label not allowed in ENTRY") if(label());
		if(symbol($f3))
		{
		}
		else
		{
			error("Invalid symbol in operand field");
		}
		next;
	}
	if($f2 eq 'EJECT')
	{
		error("Label not allowed in EJECT") if(label());
		next;
	}
	if($f2 eq 'AORG')
	{
		$nonid=1;
		error("Label not allowed in AORG") if(label());
		error("Operand missing in AORG") if($f3 eq "");
		$p = evalexpr($f3);
		if($DEBUG)
		{
			print "AORG:" . prhex(16,$p);
			print "\n";
		}
		$mode = "A";
		next;
	}
	if($f2 eq 'RORG')
	{
		$nonid=1;
		error("Label not allowed in RORG") if(label());
		error("Operand missing in AORG") if($f3 eq "");
        $p = evalexpr($f3);
		if($DEBUG)
		{
			print "RORG:" . prhex(16,$p);
			print "\n";
		}
		$mode = "R";	
		next;
	}
	if($f2 eq 'EXTRN')
	{
		$nonid=1;
		error("Label not allowed in EXTRN") if(label());
		if(symbol($f3))
		{
			if($sym{$f3} eq undef)
			{
				$sym{$f3} = 0; 		         # default value of external symbol
				$symtype{$f3} = "X"; 
			}
			else
			{
				error("External contains a double defined symbol $f3");
			}
		}
		else
		{
			error("Invalid symbol in operand field");
		}
		next;
	}
	if($f2 eq 'END')
	{
		$nonid=1;
		error("Label not allowed in END") if(label());
		last;
	}	
	
    # A label is allowed/required for the following directives	
	if($f2 eq 'RES')
	{
		$nonid=1;
		label();
		if($f3 ne "")
		{
			$l=evalexpr($f3);
			$p=$p+2*$l;
		}
		else
		{
			error("Operand missing in RES");
		}
		next;
	}
	if($f2 eq 'DATA')
	{
		$nonid=1;
		label();
		if($f3 ne "")
		{
			if(substr($f3,0,1) eq '\'')
			{
				# DATA string
				# $f3 is may be wrong because of line split in fixed length colomns
				# redo $f3 as variable length field 
				$f3 = substr($line,$tab3,1);   # '
				$l= length($line) - $tab3;     # remaining chars in line for f3
				$les=1;
				#(string length max 48 characters (50 - 2 for the quotes)
				while($les<50 && $les<=$l)
				{
					$f3=$f3 . substr($line,$tab3+$les,1);
					if(substr($line,$tab3+$les,1) eq '\'')
					{
						last;
					}
					$les=$les+1;
				}
				$l=length($f3) - 2;          # string length in characters
				if($l & 1)
				{
					# uneven number of characters, make even
					$l=$l+1;
				}
				$p=$p+$l;
			}
			else
			{
				# 1 word DATA specification
				$p=$p+2;
			}
		}
		else
		{
			error("Operand missing in DATA");
		}
		next;
	}
	if($f2 eq 'EQU')
	{
		$nonid=1;
		if(! ($symbol=label()))
		{
			error("Label missing, or wrong label in EQU, ");
		}
		else
		{
			if($f3 ne "")
			{
				if(symbol($f3) && ($symtype{$f3} eq "Z"))   # register name?
				{
					$sym{$symbol} = $sym{$f3};  # define alternative register name
					$symtype{$symbol} = "Z";    # and set type
				}
				else
				{
					$l=evalexpr($f3);
					$sym{$symbol} = $l; 		# define symbol value
					$symtype{$symbol} = $etype; # and set type
					if($etype eq "X")
					{
						error("External not allowed in EQU");
					}
				}
			}
			else
			{
				error("Operand missing in EQU");
			}
		}
		next;
	}
	
	# Scan the non-directives operations
	$nonid=1;
	label();
	if(($l=$L{$f2}) eq undef)	# $l = instruction length in words
	{
		error("Undefined opcode");
		next;
	}
	$p=$p+2*$l;
	
}

# pass1 postprocessing
if($ident == 0)
{
	error("No valid IDENT statement specified");
}

# end of pass1

close (IN) || die "can't close $fi: $!";
print "End of Assembler pass 1, number of detected errors = $nerr\n";
if($nerr != 0)
{
	print "Stop, no object file generated\n";
	exit 1;
}

###########################      PASS 2        ##########################
if($DEBUG) {print "Start reading $fi (pass2)\n";}
open (IN,$fi) || die "cannot open inputfile $fi: $!";	
open (BIN, ">:raw", $fo) || die "cannot open outputfile $fo: $!";
open (LIST,">$fl") || die "cannot open outputfile $fl: $!";
$ln=0;					# source input line number
$prl=0;                 # program length
$pass=2;
################# list page format ##############
my $page = 1;			# listing page number
my $lpmax = 40;         # lines per listing page
my $lp = 0;             # listing page line number
#################################################
$p=0;                   # reset the location counter

while(<IN>)
{
	chomp;
	$line = $_;
	if(ord(substr($_,0,1)) == 0)
	{
		# strip leader
		$i=0;
		$l=length($_);
		while((ord(substr($_,$i,1)) == 0) && $i < $l)
		{
			$i=$i+1;
		}
		$line = substr($_,$i,$l-$i);
	}	
	$ln=$ln+1;
    if (substr($line,0,1) eq '*')
	{
		# list comment line
	    list($ln, spaces(14) . $line);
	    next;
	}
    $f1=substr($line,$tab1,$fl1);        # label
	$f1=rtrim($f1);
    $f2=substr($line,$tab2,$fl2);        # operation
	$f2=rtrim($f2);
	$f2=~ tr/\.//d;                      # deleted dots from opcode
    $f3=substr($line,$tab3,$fl3);        # operand	
	$f3=rtrim($f3);
    $f4=substr($line,$tab4,$fl4);        # comment

	if($f2 eq 'IDENT')
	{
	    list($ln, spaces(14) . $line);
		gentype0($f3);                   # output IDENT block
		next;
	}
	if($f2 eq 'AORG')
	{
		$p = evalexpr($f3);
	    list($ln, spaces(14) . $line);
		$mode = "A";
		next;
	}
	if($f2 eq 'RORG')
	{
		$p = evalexpr($f3);
	    list($ln, spaces(14) . $line);
		$mode = "R";	
		next;
	}
	if($f2 eq 'EXTRN')
	{
	    list($ln, spaces(14) . $line);
		next;
	}
	if($f2 eq 'ENTRY')
	{
		if($sym{$f3} eq undef)
		{
			error("Undefined entry point")
		}
		else
		{
			gentype1($f3, $sym{$f3}, $symtype{$f3}); # output ENTRY block
		}
		list($ln, spaces(14) . $line);
		next;
	}
	if($f2 eq 'EJECT')
	{
		while($lp <= $lpmax)
		{
			print LIST "\n";
			$lp = $lp + 1;
		}
		$lp=0;
		list($ln, spaces(14) . $line);
		next;
	}	
	if($f2 eq 'END')
	{
		if($f3 ne "")
		{
			if($sym{$f3} eq undef)
			{
				error("Undefined start address used");
				gentype7(0,"0",$prl,$nerr);  # address type 0 means NO STARTADDRESS
			}
			else
			{
				print "END:S=$f3 V=$sym{$f3} T=$symtype{$f3}\n";
				gentype7($sym{$f3},$symtype{$f3},$prl,$nerr);
			}
		}
		else
		{
			gentype7(0,"0",$prl,$nerr);      # address type 0 means NO STARTADDRESS
		}
		list($ln, spaces(14) . $line);
		last;
	}	
	
    # A label is allowed/required for the following directives	
	if($f2 eq 'RES')
	{
		my $i = 0;
		
		$l=evalexpr($f3);                    # number of words to reserve
		
		list($ln, prhex(16,$p) . spaces(10) . $line);
		
		# not required, but reserved space is initialized to zero
		while($i < $l)
		{
			gentype3(($p+ ($i*2)),$mode, 0 ,"A");    # GENERATE ZERO DATA WORD
		    $i = $i + 1;
		}
		$p=$p+2*$l;
		next;
	}
	if($f2 eq 'DATA')
	{
		# only a 1 word DATA specification is allowed (at this moment) except for strings
		if(substr($f3,0,1) eq '\'')
		{
			my $c=0;
			# multiword DATA string (max 48 char!)
			# $f3 is may be wrong because of line split in fixed length colomns
			# redo $f3 as variable length field 
			$f3 = substr($line,$tab3,1);   # '
			$l= length($line) - $tab3;     # remaining chars in line for f3
			$les=1;	
			# (string length max 48 characters (50 - 2 for the quotes))
			while($les<50 && $les<=$l)
			{
				$f3=$f3 . substr($line,$tab3+$les,1);
				if(substr($line,$tab3+$les,1) eq '\'')
				{
					last;
				}
				$les=$les+1;
			}			
			$les = length($f3) - 1;
			if(substr($f3,$les,1) eq '\'' && $les > 1)
			{
				$les = $les - 1;                  # string length
				if($les == 1)
				{
					$c=substr($f3,1,1);
					if($ch{$c} eq undef)
					{
						error("Wrong character in string");
						$c=0;
					}
					else
					{
						$c=$ch{$c} << 8;
					}
					list($ln, prhex(16,$p) . "  " . prhex(16,$c) . " A  " . $line);
				    gentype3($p,$mode,$c,"A");
					$p=$p+2;
				}
				if($les >= 2)
				{
					$c=substr($f3,1,1);
					if(($c=$ch{$c}) eq undef)
					{
						error("Wrong character in string");
						$c=0;
					}
					$eval = $c << 8;
					$c=substr($f3,2,1);			
					if(($c=$ch{$c}) eq undef)
					{
						error("Wrong character in string");
						$c=0;
					}
                    $eval = $eval | $c;
					list($ln, prhex(16,$p) . "  " . prhex(16,$eval) . " A  " . $line);
				    gentype3($p,$mode,$eval,"A");
					$p=$p+2;
				}				
				if($les > 2)
				{
				    $ei=3;
					while($ei <= $les)
					{
						$c=substr($f3,$ei,1);
						if(($c=$ch{$c}) eq undef)
						{
							error("Wrong character in string");
							$c=0;
						}
						$eval = $c << 8;
						$ei=$ei+1;
						if($ei<=$les)
						{
							$c=substr($f3,$ei,1);			
							if(($c=$ch{$c}) eq undef)
							{
								error("Wrong character in string");
								$c=0;
							}							
						}
						else
						{
							$c=0;
						}
						$eval = $eval | $c;
						$ei=$ei+1;						
						list(0, spaces(9) . prhex(16,$p) . "  " . prhex(16,$eval) . " A  ");
						gentype3($p,$mode,$eval,"A");
						$p=$p+2;
					}
				}

			}
			else
			{
				error("Malformed string: quote missing, string too long, or empty string");
				list($ln, prhex(16,$p) . "  " . prhex(16,$c) . " A  " . $line);
				gentype3($p,$mode,$c,"A");	
				$les = $les - 1; 				
				if($les & 1)
				{
					$les = $les + 1;
				}
				$p=$p+$les;								
			}
			next;
		}
		else
		{
			# 1 word DATA specification
			$l=evalexpr($f3);                 # content
			list($ln, prhex(16,$p) . "  " . prhex(16,$l) . " $etype  " . $line);
			if($etype eq "X")
			{
				if($DEBUG) {print "Solve external at ".prhex(16,$p)." :".$exsym."\n";}
				gentype3($p,$mode,$l,"A");    # GENERATE ABS DATA WORD
				gentype4($exsym,$p,$mode);    # PLUS AN EXTERNAL MODIFIER
			}
			else
			{
				gentype3($p,$mode,$l,$etype); # GENERATE DATA WORD
			}
			$p=$p+2;
			next;
		}
	}
	if($f2 eq 'EQU')
	{
        list($ln, spaces(14) . $line);
		next;
	}
    # Scan machine operations
	scan($f2,$I{$f2},$f3);   # $f2=opcode, $I{$f2}=template, $f3=string
	# and generate the code
	my $w1;
	my $w2;
	my $w3;
	my $st;
	
	$w1=gen1($G1{$f2});                 # generate word 1
	
    list($ln, prhex(16,$p) . spaces(2). prhex(16,$w1) . spaces(4) . $line);	
	gentype3($p,$mode,$w1,"0");         # GENERATE CODE WORD
	$p = $p + 2;
	
	# 2 words
    if(	$L{$f2} == 2 )
	{
		$w2=gen2($G2{$f2});             # generate word 2
		
		$st = " $oty  ";
		if($oty eq "X")
		{
			if($DEBUG) {print "Solve external at ".prhex(16,$p)." :".$exsym."\n";}
			gentype3($p,$mode,$w2,"A");    # GENERATE ABS OPERAND WORD
			gentype4($exsym,$p,$mode);     # PLUS AN EXTERNAL MODIFIER	
		}
		else
		{
			gentype3($p,$mode,$w2,$oty);   # GENERATE OPERAND WORD
		}
		
		list(0, spaces(9) . prhex(16,$p) . spaces(2). prhex(16,$w2) . $st);	
		$p = $p + 2;
	}
	
	# 3 words ?
    if(	($l=$L{$f2}) == 3 )
	{
		$w2=gen2($G2{$f2});             # generate word 2
		$st = " $oty1  ";
		if($oty1 eq "X")
		{
			if($DEBUG) {print "Solve external at ".prhex(16,$p)." :".$exsym1."\n";}
			gentype3($p,$mode,$w2,"A");    # GENERATE ABS OPERAND WORD
			gentype4($exsym1,$p,$mode);    # PLUS AN EXTERNAL MODIFIER	
		}
		else
		{
			gentype3($p,$mode,$w2,$oty1);   # GENERATE OPERAND WORD
		}
		
		list(0, spaces(9) . prhex(16,$p) . spaces(2). prhex(16,$w2) . $st);	
		$p = $p + 2;
		
		$w3=gen3($G3{$f2});             # generate word 3
		$st = " $oty2  ";
		if($oty2 eq "X")
		{
			if($DEBUG) {print "Solve external at ".prhex(16,$p)." :".$exsym2."\n";}
			gentype3($p,$mode,$w3,"A");    # GENERATE ABS OPERAND WORD
			gentype4($exsym2,$p,$mode);    # PLUS AN EXTERNAL MODIFIER	
		}
		else
		{
			gentype3($p,$mode,$w3,$oty2);   # GENERATE OPERAND WORD
		}
		
		list(0, spaces(9) . prhex(16,$p) . spaces(2). prhex(16,$w3) . $st);	
		$p = $p + 2;
	}
}	

# force an EJECT to have the symbol table on a new page
while($lp <= $lpmax)
{
    print LIST "\n";
	$lp=$lp+1;
}
$lp=0;   
list(0, "  SYMBOL TABLE\n");

my $sb;
my $sl="";
$i=0;
foreach my $n (sort keys %sym)
{
	$l = length($n);
	$sb = "  $n" . spaces($symlength-$l+1) . prhex(16,$sym{$n}) . " " . $symtype{$n};
	$sl = $sl . $sb . " ";
	$i = $i + 1;
	if($i == 4)		# symbol table over 5 columns
	{
		list(0, $sl);
		$sl = "";
		$i=0;
	}
	
}
list(0, $sl); # last line symbol table
list(0, "\n  P8ASM, version $version. End of assembly, $nerr errors.");

if(! $conlist)
{
	print  "P8ASM, version $version. End of assembly, $nerr errors.\n"
}

close (IN)   || die "can't close $fi: $!";
close (BIN)  || die "can't close $fo: $!";
close (LIST) || die "can't close $fl: $!";

if($nerr == 0)
{
	exit;
}
else
{
	exit 1;
}
##############################################################################
sub punch {
    my($a) = @_;
    my $byte;
    $byte = pack("C", $a);
    print BIN $byte;
}

sub rtrim {
	my $s = shift;
	my $l = length($s);
	my $i = $l;
	if ($l > 0)
	{
		while($i > 0)
		{
			if((substr($s,$i-1,1)) ne " ")
			{ 
				last;
			}
			else
			{
				$i=$i-1;
			}
		}
	}
	else
	{
		return "";
	}
	if($i == 0)
	{
		return "";
	}
	else
	{
		return substr($s,0,$i);
	}
}

sub prhex {
	# print hex
	# $n is number of bits
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	if($n == 0)
	{
		return "";
	}
	elsif($n<5)
	{
		$s = sprintf("%01X", $v);	
	}
	elsif($n<9)
	{
		$s = sprintf("%02X", $v);
	}
	elsif($n<13)
	{		
		$s = sprintf("%03X", $v);
	}
	elsif($n<17)
	{		
		$s = sprintf("%04X", $v);
	}
	elsif($n<21)
	{
		$s = sprintf("%05X", $v);
	}
	elsif($n<25)
	{						
		$s = sprintf("%06X", $v);
	}
	elsif($n<29)
	{		
		$s = sprintf("%07X", $v);
	}
	elsif($n<33)
	{
		$s = sprintf("%08X", $v);
	}
	# print "$s ";
	return $s;
}

sub prdec
{
	# print number
	# $n is number of digits (string)
	# $v is value
	my $n = shift;
	my $v = shift;
	my $s;
	my $format="\%0" . $n . "d";
	$s = sprintf($format, $v);
	# print "$s ";
	return "  $s  ";
}

sub error
{
	my $es = shift;
	$nerr = $nerr + 1;
	if($pass == 1)
	{
		print prdec("4", $ln) . "$line    ****ERROR**** $es\n";
	}
	else
	{
		list(0, spaces(22) . "*LINE " . prdec("4", $ln) . " ********ERROR******** $es *");
	}	
}

sub label
{
	my $s;
	return "" if $f1 eq "";
	if($s=symbol($f1))
	{
		if($sym{$s} eq undef)
		{
			$sym{$s} = $p;           # default value of symbol = location address
			$symtype{$s} = $mode; 
			return $s;
		}
		else
		{
			error("ERROR Label: Label field contains a double defined symbol $s ,or is not allowed");
		}
	}
	else
	{
		error("ERROR Label: Label $f1 not a valid symbol, or is not allowed");
	}
	return "";
}

# scan whether $s contains a valid symbol
# 1st char must be alpha or _
# next chars alpha|digit|:|. (dots will be deleted)
sub symbol
{
	my $s=shift;
	my $l = length($s);
	my $c;
	my $i;
	my $result="";
	if($l > $symlength)
	{
		return "";
	}
	$i=0;
	if($l>=1) # minimal 1 character
	{
		$c=substr($s,0,1);           # get 1st character
		if($c =~ /[A-Z_]/)
		{
			# 1st character OK
			$result=$result . $c;
			$i=1;
			while($i < $l)
			{
				$c=substr($s,$i,1); # next character
				$i=$i+1;
				if($c eq '.')
				{
					next;			# delete dots
				}
				if($c =~ /[A-Z0-9:]/)   #/[A-Z0-9:_]/ to allow underscores within the symbol
				{
					$result=$result . $c;
				}
				else
				{
					return "";      # not a valid symbol
				}
			}
			return $result;
		}
	}
	return "";  # not a valid symbol
}
	

# output listing record
sub list
{
	my $ln = shift;			# source input record number
	my $ls = shift;			# string to list
	my $l=0;
	my $s;
	
	# new page ? output listing header on the new page 
	if($lp == 0)
	{
		if($conlist) {print "\n\n\n";}
		print LIST "\n\n\n";
		$l = length($header);
		$s = spaces(23) . $header . spaces(88-22-$l) . "PAGE" . prdec(4,$page);
		if($conlist) {print $s;}
		print LIST $s;
		if($conlist) {print "\n\n\n";}
		print LIST "\n\n\n";
		$page = $page + 1;			# page number
		$lp = 3;
	}
	if($ln == 0)
	{
		# just output the string
		$s = $ls;
	}
	else
	{
		# output an assembled record
		# $s = prdec(4,$ln) . spaces(14) . $ls;
		$s = prdec(5,$ln) . $ls;
	}
	if($conlist) {print $s;}
	print LIST $s;
	if($conlist) {print "\n";}
	print LIST "\n";
    $lp = $lp + 1;
	if($lp == $lpmax)
	{
		$lp = 0;			# reset internal line counter to force a new page
	}
}

# generate a string of n spaces
sub spaces
{
	my $n = shift;
	my $i = 0;
	my $s = "";
	while($i < $n)
	{
		$s = $s . " ";
		$i = $i + 1;
	}
	return $s;
}

######################### expression evaluator ###############################
# provide next character of the expression string
sub nextchar
{
	my $c;
	if($ei <= $les)
	{
		$c = substr($es,$ei,1);
		$ei = $ei + 1;
		return $c;
	}
	else
	{
		return " ";
	}
}

# peek next character of the expression string
sub peekchar
{
	my $c;
	if($ei <= $les)
	{
		$c = substr($es,$ei,1);
		return $c;
	}
	else
	{
		return " ";
	}
	
}

# peek next-next character of the expression string
sub peekchar2
{
	my $c;
	if(($ei+1) <= $les)
	{
		$c = substr($es,$ei+1,1);
		return $c;
	}
	else
	{
		return " ";
	}
	
}

# convert hex string in expression to binary
sub cvhex
{
	my $r=0;
	my $d=0;
	my $c;
	if( ($c = peekchar()) =~ /[0-9A-F]/ )	# at least 1 hex digit required
	{
		while (peekchar() =~ /[0-9A-F]/)
		{
			$c = nextchar();
			$d = $cvdig{$c};
			$r = ($r << 4) + $d;
		}
		if ($r > 0xFFFF)
		{
			$exerr = 1;
			error("Error in hex number (too big for 16 bit)");
			return 0;
		}
		return ($r & 0xFFFF);
	}
	else
	{
		$exerr = 1;
		error("Error in hex number");
		return 0;
	}
}

# convert decimal string in expression to binary
sub cvdec
{
	my $r=0;
	my $c;
	if( ($c = peekchar()) =~ /[0-9]/ )	# at least digit required
	{
		while ( (peekchar()) =~ /[0-9]/)
		{
			$c = nextchar();
			$c = $cvdig{$c};
			$r = $r * 10 + $c;
		}
		if ($r > 0xFFFF)
		{
			$exerr = 1;
			error("Error in decimal number (too big for 16 bit)");
			return 0;
		}
		return ($r & 0xFFFF);
	}
	else
	{
		$exerr = 1;
		error("Error in decimal number");
		return 0;
	}
}

# Calculates a 16bit absolute, relocatable, external value for an expression
# in the operand field (field3)

# Register expressions can only have the simple form:
# LABEL  EQU A1
# NLABEL EQU LABEL
# so a undefined symbol can be equated to a predefined or defined register symbol only
# only 1 term (a predefined or defined register symbol) is allowed in the operand field
# these type of expressions are directly handled in the EQU code of pass 1

# expr =:: 'xx' (string expression)
# expr =:: @(+|-)term@(+|-term)...
# Expression is an optional + or - followed by a term optionally followed by +
# or - another term followed by + or - another term, and so on, where term is:
# - an absolute constant, or
# - an absolute symbolic value, or
# - a relocational symbolic value, or
# - an external symbolic value.
#	(* represents either the absolute or relocatable symbolic value of the
#    location counter depending on AORG or RORG mode)
#
# However to keep the assembler and loader simple, the following conditions
# apply:
# 1) Only a single positive external symbol is allowed in an expression,
#	and if each absolute and external term in the expression is replaced by a
#	0 and each relocatable term is replace by a 1, then
# 2) If only absolute and relocatable terms are present in the expression,
#	then the expression must evaluate to 1;
# 3) If the positive external is present in the expression together with
#	absolute and relocatable terms, the the expression must evaluate to 0.
#	In all other cases (result smaller than 0 or bigger than 1) the expression
#	is malformed.
#
#	When the expression evaluates to 1, the value is relocatable
#	When the expression evaluates to 0, without the external being present,
#	the value is absolute.
#	When the external is present the value is external.
#
sub evalexpr
{
	my $c;
	my $sign = '+';
	my $systr = "";
	my $r;
	my $x;
	
	$es = shift;			 # expression string
	$les = length($es);      # length expression string
	$ei = 0;                 # char index in expression string
    $eval = 0;               # result
    $etype = "";             # A = Absolute, R = Relocatable, X = External, "!" = Error
	$exerr = 0;              # reset error flag
	$exsym = "";             # external symbol 
	$r = 0;					 # validity check
	$x = 0;					 # number of scanned externals in expression
	
	#print "expr:$es\n";
	if($les <= 0)
	{
		$exerr = 1;
		error("Expected expression is missing in operand");
	}
	while( $ei < $les)
	{
		$systr = "";
		$c = peekchar();
		
		if( $c eq '*' )
		{
			nextchar();
			if($sign eq '+')
			{
				$eval = $eval + $p;
				if($mode eq "R")
				{
					$r =$r + 1;
				}
			}
			else
			{
				$eval = $eval - $p;
				if($mode eq "R")
				{
					$r =$r - 1;
				}
			}
		    next;
		}
		elsif( $c eq '/')
		{
			nextchar();
			if($sign eq '+')
			{
				$eval = $eval + cvhex();
			}
			else
			{
				$eval = $eval - cvhex();
			}
			next;
		}
		elsif(($c eq 'X') && (peekchar2() eq "\'"))
		{
			nextchar(); # skip X
			nextchar(); # skip '
			$eval = $eval + cvhex();
			$c = nextchar();
			if($c ne "\'")
			{
				$exerr = 1;
				error("Error hex string, missing \' at the end of the string");
			}
			last;
		}
		elsif( $c =~ /[0-9]/)
		{
			if($sign eq '+')
			{
				$eval = $eval + cvdec();
			}
			else
			{
				$eval = $eval - cvdec();
			}			
			next;
		}
		elsif( $c =~ /[A-Z_]/)
		{
			# symbol
			while( peekchar() =~ /[A-Z0-9_:\.]/ )
			{
				$c = nextchar();
				if($c eq '.')  # skip dots
				{
					next;
				}
				$systr = $systr . $c;
			}
			if(length($systr) > $symlength)
			{
				$exerr = 1;
				error("Symbol $systr in operand longer than $symlength characters");
				last;
			}
			if($sym{$systr} ne undef)
			{
				if($sign eq '+')
				{
					$eval = $eval + $sym{$systr};
					if($symtype{$systr} eq "R")
					{
						$r =$r + 1;
					}
					if($symtype{$systr} eq "X")
					{
						$x = $x + 1;
						$exsym = $systr;		# save external symbol
					}
					if($symtype{$systr} eq "Z")
					{
						$exerr = 1;
						error("Register $systr is not allowed in operand expression");
						last;
				    }
				}
				if($sign eq '-')
				{
					$eval = $eval - $sym{$systr};
					if($symtype{$systr} eq "R")
					{
						$r =$r - 1;
					}
					if($symtype{$systr} eq "X")
					{
						$exerr = 1;
						error("Only positive external values are allowed");
						last;
					}
					if($symtype{$systr} eq "Z")
					{
						$exerr = 1;
						error("Register $systr is not allowed in operand expression");
						last;
				    }					
				}
				next;
			}
			else
			{
				$exerr = 1;
				error("Undefined symbol $systr in operand expression");
				last;
			}
		}
		elsif( $c eq '+' || $c eq '-' )
		{
			$sign = nextchar();
			next;
		}
		elsif( $c eq '\'' && $ei == 0 && $les < 5)
		{
			# short string expression (constant or long constant in instruction)
			# string expressions in the DATA statement are processed by the DATA code
			# always an absolute constant
			if(substr($es,2,1) eq '\'')
			{
				$eval=$ch{substr($es,1,1)};
			    last;
			}
			elsif(substr($es,3,1) eq '\'')
			{
				$eval=$ch{substr($es,1,1)};
				$eval = $eval << 8;
				$eval = $eval + $ch{substr($es,2,1)};
				last;
			}
			else
			{
				$exerr = 1;
				error("Wrong string or string bigger than 1 or 2 characters");
				last;
			}
		}
		else
		{
			$exerr = 1;
			error("Operand expression error");
			last;
		}
	}
	if($r == 1 && $x == 0)
	{
		$etype = 'R';
	}
	elsif($r == 0 && $x == 0)
	{
		$etype = 'A';
	}
	elsif($r == 0 && $x == 1)
	{
		$etype = 'X';
	}
	else
	{
		$exerr = 1;
		error("Malformed operand expression");
	}
	
	if($exerr)
	{
		$etype = '!';
	}
	return ($eval = ($eval & 0xFFFF));
}

########################### operand syntax scanner #########################
sub scan
{
	my $o = shift;           # operation code
	my $t = shift;           # operand template
	my $s = shift;           # operand to scan
	
	my $t1;                  # t1,t2,t3 template parts
	my $t2;
	my $t3;
	
	my $p1;                  # p1,p2,p3 operand parts
	my $p2;
	my $p3;

	my $l;
	
	($t1,$t2,$t3) = split(/\,/, $t);  # split template
	($p1,$p2,$p3) = split(/\,/, $s);  # split operand
	$oty = '';
	# scan part 1
	$l = length($t1);
	if($l>0)
    {
		if($t1 eq 'r1')
		{
			if($sym{$p1} eq undef || $symtype{$p1} ne "Z")
			{
				error("Wrong or no register used");
			}
			else
			{
				$r1 = $sym{$p1};
			}
		}
		elsif($t1 eq 'r2')
		{
			if($sym{$p1} eq undef || $symtype{$p1} ne "Z")
			{
				error("Wrong or no register used");
			}
			else
			{
				$r2 = $sym{$p1};
			}
		}
		elsif($t1 eq 'r3')
		{
			if(($sym{$p1} eq undef) || ($symtype{$p1} ne "Z") || ($sym{$p1} & 1))
			{
				error("Wrong or no register used");
			}
			else
			{
				$r3 = $sym{$p1} >> 1;
			}
		}
		elsif($t1 eq 'm')
		{
			$m = evalexpr($p1);
			$oty = $etype;
			if($m & 0x1)
			{
				error("Uneven address in operand");
			}
			
		}
		elsif($t1 eq 'nr')
		{
			$nr = evalexpr($p1);
			if($nr >= 16)
			{
				error("Register count bigger than 15 (4 bit)");
			}
			$nr = $nr & 0xF;
		}
		elsif($t1 eq 'ns')
		{
			$ns = evalexpr($p1);
			if($ns >= 32)
			{
				error("Shift count bigger than 31 (5 bit)");
			}
			$ns = $ns & 0x1F;
		}
		elsif($t1 eq 'D')
		{
			$D = evalexpr($p1);
			if($o =~ /RB/)
			{
				$D = $p + 2 - $D;
				if($D < 2 || $D >= 256)
				{
					error ("Wrong displament in RB ". prhex(16,$D) . " bytes");
				}
			}
			elsif($o =~ /RF/)
			{
				$D = $D - $p - 2;
				if($D < 0 || $D >= 256)
				{
					error ("Wrong displament in RF ". prhex(16,$D) . " bytes");
				}
			}
			elsif($o =~ /AB/)
			{
				# for AB.., D is an absolute address between 0 and 255
				if($D < 0 || $D >= 256 || $etype ne 'A')
				{
					error ("Wrong absolute address in AB ". prhex(16,$D));
				}
			}			
			else
			{
				error ("Assembler error: D requested without RB, RF or AB");
			}
			$D=($D >> 1) & 0x7F;
		}
		elsif($t1 eq 'k')
		{
			$k = evalexpr($p1);
			if($k >= 256 || $etype ne 'A')
			{
				error ("Short constant not abs or bigger than 256 (8 bit)");
			}
			$k = $k & 0xFF;
		}
		elsif($t1 eq 'ik')
		{
			$ik = evalexpr($p1);
			
			if($ik >= 32)
			{
				error ("Internal interrupt number bigger than 31 (5 bit)");
			}
			$ik = $ik & 0x1F;
		}
		elsif($t1 eq 'lk')
		{
			$lk = evalexpr($p1);
			$oty = $etype;
		}
		elsif($t1 eq 'lk1')
		{
			$lk = evalexpr($p1);
			$oty1 = $etype;
			$exsym1 = $exsym;
		}
		else
		{
			error("Operand expected");
		}
	}
	else
	{
		if($p1 eq '')
		{
			# instruction without operand
			
		}
		else
		{
			error("No operand expected");
		}
		return;
	}

	# scan part 2
	$l = length($t2);
	if($l>0)
    {
		if($t2 eq 'r2')
		{
			if($sym{$p2} eq undef || $symtype{$p2} ne "Z")
			{
				error("Wrong or no register used");
			}
			else
			{
				$r2 = $sym{$p2};
			}				
		}
		elsif($t2 eq '@r2')
		{
			if($p2 ne "")
			{
				# option available
				if($sym{$p2} eq undef || $symtype{$p2} ne "Z")
				{
					error("Wrong or no register used");
				}
				else
				{
					$r2 = $sym{$p2};
				}					
			}
			else
			{
				# option not available
				$r2 = 0;
			}
		}
		elsif($t2 eq 'm')
		{
			$m = evalexpr($p2);
			$oty = $etype;
			#if($m & 0x1)
			#{
			#	error("Uneven address in operand");
			#}
		}
		elsif($t2 eq 'ns')
		{
			$ns = evalexpr($p2);
			if($ns >= 32)
			{
				error ("Shift count bigger than 31 \(5 bit\)");
			}
			$ns = $ns & 0x1F;
		}
		elsif($t2 eq 'k')
		{
			$k = evalexpr($p2);			
			if($k >= 256 || $etype ne 'A')
			{
				error ("Short constant not ABS or bigger than 256 \(8 bit\)");
			}
			$k = $k & 0xFF;
		}
		elsif($t2 eq 'lk')
		{
			$lk = evalexpr($p2);
			$oty = $etype;
		}
		elsif($t2 eq 'lk2')
		{
			$lk2  = evalexpr($p2);
			$oty2 = $etype;
			$exsym2 = $exsym;
		}
		elsif($t2 eq 'df')
		{
			$df = evalexpr($p2);
			if($df != 0 && $df != 1) 
			{
				error("IO operation requires a single bit function");
			    $df = 0;
			}
		}
		elsif($t2 eq 'dn')
		{
			$dn = evalexpr($p2);
			if($dn >= 64)
			{
				error ("Device address bigger than 64 \(6 bit\)");
			}
			$dn = $dn & 0x3F;
		}
		else
		{
			error("2nd operand field expected ($t2)");
		}
	}
	else
	{
		if($p2 ne '')
		{
			error("No 2nd operand field expected")
		}
		return;
	}
	
	# scan part 3
	$l = length($t3);
	if($l>0)
    {
		if($t3 eq '@r2')
		{
			if($p3 ne "")
			{
				# option available
				if($sym{$p3} eq undef || $symtype{$p3} ne "Z")
				{
					error("Wrong or no register used");
				}
				else
				{
					$r2 = $sym{$p3};
				}				
			}
			else
			{
				# option not available
				$r2 = 0;
			}
		}
		elsif($t3 eq 'dn')
		{
			$dn = evalexpr($p3);
			if($dn >= 64)
			{
				error ("Device address bigger than 64 \(6 bit\)");
			}
			$dn = $dn & 0x3F;
		}
		else
		{
			error("3rd operand field expected ($t3)");
		}
	}
	else
	{
		if($p3 ne '')
		{
			error("No 3rd operand field expected");
		}
		return;
	}

}

####################### code generator ##########################
# generator of word 1
sub gen1 {
	my $s = shift;
	my $c1;
	my $c2;
	my $c3;
	my $c4;
	my $c5;
	my $w=0;
	my $l;
	my $i;
	my $c;
	my $b;
	($c1,$c2,$c3,$c4,$c5) = split(/\./, $s);
	
	# part 1
	$l = length($c1);
	if($l>0)
    {
		for($i=0; $i<$l; $i++)
		{
			$c=substr($c1,$i,1);
			if($c eq '0')
			{
				$b = 0;
			}
			else
			{
				$b = 1;
			}
			$w = $w << 1;
			$w=$w+$b;
		}
	}
	
	# part 2
	$l = length($c2);
	if($l>0)
	{
		if($c2 eq 'r1')
		{
			$w = $w << 4;
			$w = $w + $r1;
		}
		elsif($c2 eq 'r2')
		{	
			$w = $w << 4;
			$w = $w + $r2;
		}
		elsif($c2 eq 'r3')
		{	
			$w = $w << 3;
			$w = $w + $r3;
		}
		elsif($c2 eq 'nr')
		{
			$w = $w << 4;
			$w = $w + $nr;
		}
		elsif($c2 eq 'ns')
		{
			$w = $w << 5;
			$w = $w + $ns;
		}
		elsif($c2 eq 'D')
		{
			$w = $w << 7;
			$w = $w + $D;
		}
		elsif($c2 eq 'ik')
		{
			$w = $w << 5;
			$w = $w + $ik;
		}
		elsif($c2 eq 'k')
		{
			$w = $w << 8;
			$w = $w + $k;
		}
		else
		{
			error("Code gen error w1 part2 $c2");
		}
	}
	
	# part 3
	$l = length($c3);
	if($l>0)
    {
		if($c3 eq 'k')
		{
			$w = $w << 8;
			$w = $w + $k;
		}
		else
		{
			for($i=0; $i<$l; $i++)
			{
				$c=substr($c3,$i,1);
				if($c eq '0')
				{
					$b = 0;
				}
				else
				{
					$b = 1;
				}
				$w = $w << 1;
				$w=$w+$b;
			}
		}
	}
	
	# part 4
	$l = length($c4);
	if($l>0)
    {
		if($c4 eq 'r2')
		{
			$w = $w << 4;
			$w = $w +$r2;
		}
		elsif($c4 eq 'ns')
		{
			$w = $w << 5;
			$w = $w +$ns;
		}
		elsif($c4 eq 'dn')
		{
			$w = $w << 6;
			$w = $w +$dn;
		}
		elsif($c4 eq 'df')
		{
			$w = $w << 1;
			$w = $w +$df;
		}
		else
		{
			error("Code gen error w1 part4 $c4");
		}
	}
	
	# part 5
	$l = length($c5);
	if($l>0)
    {
		if($c5 eq 'dn')
		{
			$w = $w << 6;
			$w = $w +$dn;
		}
		else
		{
			for($i=0; $i<$l; $i++)
			{
				$c=substr($c5,$i,1);
				if($c eq '0')
				{
					$b = 0;
				}
				else
				{
					$b = 1;
				}
				$w = $w << 1;
				$w=$w+$b;
			}
		}
	}
	return $w;
}

# generator of word 2
sub gen2 {
	my $s = shift;
	my $c1;
	my $c2;
	my $c3;
	my $w=0;
	my $l;
	my $i;
	my $c;
	my $b;
	($c1,$c2,$c3) = split(/\./, $s);
	
	# part 1
	$l = length($c1);
	if($l>0)
	{
		if($c1 eq 'm')
		{
			$w = $m;
		}
		elsif($c1 eq 'lk')
		{	
			$w = $lk;
		}
		elsif($c1 eq 'lk1')
		{	
			$w = $lk1;
		}
		else
		{
			error("Code gen error w2 part1 $c1");
		}
	}
	return $w;
}

# generator of word 3
sub gen3 {
	my $s = shift;
	my $c1;
	my $c2;
	my $c3;
	my $w=0;
	my $l;
	my $i;
	my $c;
	my $b;
	($c1,$c2,$c3) = split(/\./, $s);
	
	# part 1
	$l = length($c1);
	if($l>0)
	{
		if($c1 eq 'lk2')
		{	
			$w = $lk2;
		}
		else
		{
			error("Code gen error w3 part1 $c1");
		}
	}
	return $w;
}

################## binary output code generator ########################

# generate block type 0 (IDENT)
#
# som.type0.NAMExxxxxx.eom (total length = 13 bytes)
sub gentype0
{
	if($abshex==0)
	{
		my $name = shift;
		my $l = length($name);
		my $bty = 0; # block type = 0
		my $i = 0;
		my $c = '';
		punch($rub);			# start new module with a rubout
		punch($som);
		punch($bty);
		while($i < $l)
		{
			$c = substr($name,$i,1);
			$c=unpack("C",$c) ;
			punch($c);
			$i = $i + 1;
		}
		while($i < 10)
		{
			punch($space);
			$i = $i + 1;
		}
		punch($eom);
	}
}

# generate block type 1 (ENTRY)
#
# som.type1.NAMExxxxxx.AD.M.eom (total length = 16 bytes)
sub gentype1
{
	if($abshex==0)
	{
		my $name = shift;
		my $addr = shift;
		my $ty   = shift;
		my $l = length($name);
		my $bty = 1; # block type = 1
		my $i = 0;
		my $c = '';
		punch($som);
		punch($bty);
		while($i < $l)
		{
			$c = substr($name,$i,1);
			$c=unpack("C",$c);
			punch($c);
			$i = $i + 1;
		}
		while($i < 10)
		{
			punch($space);
			$i = $i + 1;
		}
		$i = $addr >> 8;
		punch($i);
		$i = $addr & 0xFF;
		punch($i);
		$c=unpack("C",$ty);
		punch($c);
		punch($eom);
	}
	else
	{
		print "Error: No entry points allowed in absolute hex format\n";
		$nerr=$nerr+1;
	}
}
	
# generate block type 7 (END)
#
# som.type7.AD.M.PRL,NERR.eom (total length = 9 bytes)
sub gentype7
{
	if($abshex==0)
	{
		my $addr = shift;
		my $ty   = shift;
		my $prl  = shift;
		my $errors = shift;
		my $bty = 7; # block type = 7
		my $i = 0;
		punch($som);
		punch($bty);
		$i = $addr >> 8;     # start address
		punch($i);
		$i = $addr & 0xFF;
		punch($i);
		$i=unpack("C",$ty);  # mode start address
		punch($i);
		$i = $prl >> 8;      # program length 
		punch($i);
		$i = $prl & 0xFF;
		punch($i);
		$i = $errors & 0xFF;
		punch($i);
		punch($eom);
	}
}

# generate block type 3 (CODE/DATA)
#
# som.type3.AD.M.WD.M.eom (total length = 9 bytes)
sub gentype3
{
	my $addr = shift;
	my $aty  = shift;
	my $wrd  = shift;
	my $wty  = shift;     # A R 0
	my $bty = 3;          # block type = 3
	my $i = 0;
	if($abshex==0)
	{
		punch($som);
		punch($bty);
		$i = $addr >> 8;      # address
		punch($i);
		$i = $addr & 0xFF;
		punch($i);
		$i=unpack("C",$aty);  # mode address
		punch($i);
	
		$i = $wrd >> 8;       # word
		punch($i);
		$i = $wrd & 0xFF;
		punch($i);
		$i=unpack("C",$wty);  # mode word
		punch($i);
		punch($eom);
		$prl = $prl + 1;      # update program length in words
	}
	else
	{
		if($DEBUG){print "tya:$aty tyw:$wty\n";}
		if($aty eq "A" && ($wty eq "0" || $wty eq "A")) # code word has type 0
		{
			print BIN prhex(16,$addr) . "," . prhex(16,$wrd) . "\n";
		}
		else
		{
			print "Error: In hex output format, only abs values are allowed in the address and word field\n";
			$nerr=$nerr+1;
		}
	}
}

# generate block type 4 (EXTRN modify)
#
# som.type4.NAMExxxxxx.AD.M.eom (total length = 16 bytes)
sub gentype4
{
	if($abshex==0)
	{
		my $name = shift;
		my $addr = shift;
		my $ty   = shift;
	
		my $l = length($name);
		my $bty = 4; # block type = 4
		my $i = 0;
		my $c = '';
		punch($som);
		punch($bty);
		while($i < $l)                # external name
		{
			$c = substr($name,$i,1);
			$c=unpack("C",$c);
			punch($c);
			$i = $i + 1;
		}
		while($i < 10)
		{
			punch($space);
			$i = $i + 1;
		}
		$i = $addr >> 8;              # address were external is used
		punch($i);                    # when address of external is known,
		$i = $addr & 0xFF;            # add that address to the content of
		punch($i);                    # this address
		$c=unpack("C",$ty);
		punch($c);                    # mode
		punch($eom);
	}
	else
	{
		print "Error: No externals allowed in absolute hex output format\n";
		$nerr=$nerr+1;
	}
}
